import asyncio
import base64
import contextlib
import copy
import dataclasses
import datetime
import functools
import inspect
import io
import logging
import math
import multiprocessing
import numbers
import re
import secrets
import time
import traceback
import typing
import urllib.request
from contextvars import ContextVar
from multiprocessing.process import BaseProcess
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
    cast,
)

import pyarrow  # type: ignore

import writer.blocks
import writer.evaluator
from writer import core_ui
from writer.core_ui import Component
from writer.logs import use_stdout_redirect
from writer.ss_types import (
    BlueprintExecutionError,
    BlueprintExecutionLog,
    InstancePath,
    Readable,
    ServeMode,
    WriterEvent,
    WriterEventResult,
    WriterFileItem,
)
from writer.vault import writer_vault

if TYPE_CHECKING:
    import pandas

    from writer.app_runner import AppProcess
    from writer.blueprints import BlueprintRunner
    from writer.ss_types import AppProcessServerRequest


@dataclasses.dataclass
class CurrentRequest:
    session_id: str
    request: "AppProcessServerRequest"


_current_request: ContextVar[Optional[CurrentRequest]] = ContextVar("current_request", default=None)


@contextlib.contextmanager
def use_request_context(session_id: str, request: "AppProcessServerRequest"):
    """
    Context manager to set the current request context.

    >>> session_id = "xxxxxxxxxxxxxxxxxxxxxxxxx"
    >>> request = AppProcessServerRequest(type='event', payload=EventPayload(event='my_event'))
    >>> with use_request_context(session_id, request):
    >>>     pass
    """
    try:
        _current_request.set(CurrentRequest(session_id, request))
        yield
    finally:
        _current_request.set(None)


def get_app_process() -> "AppProcess":
    """
    Retrieves the Writer Framework process context.

    >>> _current_process = get_app_process()
    >>> _current_process.bmc_components # get the component tree
    """
    from writer.app_runner import AppProcess  # Needed during runtime

    raw_process: BaseProcess = multiprocessing.current_process()
    if isinstance(raw_process, AppProcess):
        return raw_process

    raise RuntimeError("Failed to retrieve the AppProcess: running in wrong context")


def import_failure(rvalue: Any = None):
    """
    This decorator captures the failure to load a volume and returns a value instead.

    If the import of a module fails, the decorator returns the value given as a parameter.

    >>> @import_failure(rvalue=False)
    >>> def my_handler():
    >>>     import pandas
    >>>     return pandas.DataFrame()

    :param rvalue: the value to return
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ImportError:
                return rvalue

        return wrapper

    return decorator


class Config:
    is_mail_enabled_for_log: bool = False
    mode: ServeMode = "run"
    logger: Optional[logging.Logger] = None
    feature_flags: list[str] = []


class WriterSession:
    """
    Represents a session.
    """

    def __init__(
        self, session_id: str, cookies: Optional[Dict[str, str]], headers: Optional[Dict[str, str]]
    ) -> None:
        self.session_id = session_id
        self.cookies = cookies
        self.headers = headers
        self.last_active_timestamp: int = int(time.time())
        new_state = WriterState.get_new()
        new_state.user_state.mutated = set()
        self.session_state = new_state
        self.session_component_tree = core_ui.build_session_component_tree(base_component_tree)
        self.event_handler = EventHandler(self)
        self.userinfo: Optional[dict] = None
        self.queued_messages: List[Any] = []

    def update_last_active_timestamp(self) -> None:
        self.last_active_timestamp = int(time.time())


@dataclasses.dataclass
class MutationSubscription:
    """
    Describes a subscription to a mutation.

    The path on which this subscription is subscribed and the handler
    to execute when a mutation takes place on this path.

    >>> def myhandler(state):
    >>>     state["b"] = state["a"]

    >>> m = MutationSubscription(path="a.c", handler=myhandler)
    """

    type: Literal["subscription", "property"]
    path: str
    handler: Callable  # Handler to execute when mutation happens
    state: "State"
    property_name: Optional[str] = None

    def __post_init__(self):
        if len(self.path) == 0:
            raise ValueError("path cannot be empty.")

        path_parts = parse_state_variable_expression(self.path)
        for part in path_parts:
            if len(part) == 0:
                raise ValueError(f"path {self.path} cannot have empty parts.")

    @property
    def local_path(self) -> str:
        """
        Returns the last part of the key to monitor on the state

        >>> m = MutationSubscription(path="a.c", handler=myhandler)
        >>> m.local_path
        >>> "c"
        """
        path_parts = parse_state_variable_expression(self.path)
        return path_parts[-1]


class StateRecursionWatcher:
    limit = 128

    def __init__(self):
        self.counter_recursion = 0


_state_recursion_watcher = ContextVar("state_recursion_watcher", default=StateRecursionWatcher())


@contextlib.contextmanager
def state_recursion_new(key: str):
    """
    Context manager to watch the state recursion on mutation subscriptions.

    The context throws a RecursionError exception if more than 128 cascading mutations
    are performed on the same state
    """
    recursion_watcher = _state_recursion_watcher.get()
    try:
        recursion_watcher.counter_recursion += 1
        if recursion_watcher.counter_recursion > recursion_watcher.limit:
            raise RecursionError(f"State Recursion limit reached {recursion_watcher.limit}.")
        yield
    finally:
        recursion_watcher.counter_recursion -= 1


class FileWrapper:
    """
    A wrapper for either a string pointing to a file or a file-like object with a read() method.
    Provides a method for retrieving the data as data URL.
    Allows for convenient serialisation of files.
    """

    def __init__(self, file: Union[Readable, str], mime_type: Optional[str] = None):
        if not file:
            raise ValueError("Must specify a file.")
        if not (callable(getattr(file, "read", None)) or isinstance(file, str)):
            raise ValueError(
                "File must provide a read() method or contain a string with a path to a local file."
            )
        self.file = file
        self.mime_type = mime_type

    def _get_file_stream_as_dataurl(self, f_stream: Readable) -> str:
        base64_str = base64.b64encode(f_stream.read()).decode("latin-1")
        dataurl = (
            f"data:{self.mime_type if self.mime_type is not None else ''};base64,{ base64_str }"
        )
        return dataurl

    def get_as_dataurl(self) -> str:
        if isinstance(self.file, str):
            with open(self.file, "rb") as f_stream:
                return self._get_file_stream_as_dataurl(f_stream)
        elif callable(getattr(self.file, "read", None)):
            return self._get_file_stream_as_dataurl(self.file)
        else:
            raise ValueError("Invalid file.")


class BytesWrapper:
    """
    A wrapper for raw byte data.
    Provides a method for retrieving the data as data URL.
    Allows for convenient serialisation of byte data.
    """

    def __init__(self, raw_data: Any, mime_type: Optional[str] = None):
        self.raw_data = raw_data
        self.mime_type = mime_type

    def get_as_dataurl(self) -> str:
        b64_data = base64.b64encode(self.raw_data).decode("utf-8")
        durl = f"data:{self.mime_type if self.mime_type is not None else ''};base64,{ b64_data }"
        return durl


class StateSerialiserException(ValueError):
    pass


class StateSerialiser:
    """
    Serialises user state values before sending them to the front end.
    Provides JSON-compatible values, including data URLs for binary data.
    """

    def serialise(self, v: Any) -> Union[Dict, List, str, bool, int, float, None]:
        from writer.ai import Conversation
        from writer.core_df import EditableDataFrame

        if isinstance(v, State):
            return self._serialise_dict_recursively(v.to_dict())
        if isinstance(v, Conversation):
            return v.serialized_messages
        if isinstance(v, (FileWrapper, BytesWrapper)):
            return self._serialise_ss_wrapper(v)
        if isinstance(v, (datetime.datetime, datetime.date)):
            return str(v)
        if isinstance(v, bytes):
            return self.serialise(BytesWrapper(v))
        if isinstance(v, dict):
            return self._serialise_dict_recursively(v)
        if isinstance(v, list):
            return self._serialise_list_recursively(v)
        if isinstance(v, (str, bool)):
            return v
        if isinstance(v, EditableDataFrame):
            table = v.pyarrow_table()
            return self._serialise_pyarrow_table(table)
        if v is None:
            return v

        # Checking the MRO allows to determine object type without creating dependencies
        # to these packages

        v_mro = [f"{x.__module__}.{x.__name__}" for x in inspect.getmro(type(v))]

        if isinstance(v, (int, float)):
            if "numpy.float64" in v_mro:
                return float(v)
            if math.isnan(v):
                return None
            return v

        if "pandas.core.frame.DataFrame" in v_mro:
            return self._serialise_pandas_dataframe(v)
        if hasattr(v, "__dataframe__"):
            return self._serialize_dataframe(v)

        if "matplotlib.figure.Figure" in v_mro:
            return self._serialise_matplotlib_fig(v)
        if "plotly.graph_objs._figure.Figure" in v_mro:
            return v.to_json()
        if "numpy.float64" in v_mro:
            return float(v)
        if "numpy.ndarray" in v_mro:
            return self._serialise_list_recursively(v.tolist())
        if "pyarrow.lib.Table" in v_mro:
            return self._serialise_pyarrow_table(v)

        if hasattr(v, "to_dict") and callable(v.to_dict):
            # Covers Altair charts, Plotly graphs
            return self._serialise_dict_recursively(v.to_dict())

        raise StateSerialiserException(
            f"Object of type { type(v) } (MRO: {v_mro}) cannot be serialised."
        )

    def _serialise_dict_recursively(self, d: Dict) -> Dict:
        return {str(k): self.serialise(v) for k, v in d.items()}

    def _serialise_list_recursively(self, l: List) -> List:  # noqa: E741
        return [self.serialise(v) for v in l]

    def _serialise_ss_wrapper(self, v: Union[FileWrapper, BytesWrapper]) -> str:
        return v.get_as_dataurl()

    def _serialise_matplotlib_fig(self, fig) -> str:
        # It's safe to import matplotlib here without listing it as a dependency.
        # If this method is called, it's because a matplotlib figure existed.
        # Note: matplotlib type needs to be ignored since it doesn't provide types
        import matplotlib.pyplot as plt  # type: ignore

        iobytes = io.BytesIO()
        fig.savefig(iobytes, format="png")
        iobytes.seek(0)
        plt.close(fig)
        return FileWrapper(iobytes, "image/png").get_as_dataurl()

    def _serialize_dataframe(self, df) -> str:
        """
        Serialize a dataframe with pyarrow a dataframe that implements
        the Dataframe Interchange Protocol i.e. the __dataframe__() method

        :param df: dataframe that implements Dataframe Interchange Protocol (__dataframe__ method)
        :return: a arrow file as a dataurl (application/vnd.apache.arrow.file)
        """
        import pyarrow.interchange  # type: ignore

        table = pyarrow.interchange.from_dataframe(df)
        return self._serialise_pyarrow_table(table)

    def _serialise_pandas_dataframe(self, df):
        import pyarrow as pa  # type: ignore

        pa_table = pa.Table.from_pandas(df, preserve_index=True)
        return self._serialise_pyarrow_table(pa_table)

    def _serialise_pyarrow_table(self, table):
        import pyarrow as pa  # type: ignore

        sink = pa.BufferOutputStream()
        batches = table.to_batches()
        with pa.ipc.new_file(sink, table.schema) as writer:
            for batch in batches:
                writer.write_batch(batch)
        buf = sink.getvalue()
        bw = BytesWrapper(buf, "application/vnd.apache.arrow.file")
        return self.serialise(bw)


class MutableValue:
    """
    MutableValue allows you to implement a value whose modification
    will be followed by the state of Writer Framework and will trigger the refresh
    of the user interface.

    >>> class MyValue(MutableValue):
    >>>     def __init__(self, value):
    >>>         self.value = value
    >>>
    >>>     def modify(self, new_value):
    >>>         self.value = new_value
    >>>         self.mutate()
    """

    def __init__(self):
        self._mutated = False

    def mutated(self) -> bool:
        """
        Returns whether the value has been mutated.
        :return:
        """
        return self._mutated

    def mutate(self) -> None:
        """
        Marks the value as mutated.
        This will trigger the refresh of the user interface on the next round trip
        :return:
        """
        self._mutated = True

    def reset_mutation(self) -> None:
        """
        Resets the mutation flag to False.
        :return:
        """
        self._mutated = False


class StateProxy:
    """
    The root user state and its children (nested states) are instances of this class.
    Provides proxy functionality to detect state mutations via assignment.
    """

    def __init__(self, raw_state: Dict = {}):
        self.state: Dict[str, Any] = {}
        self.local_mutation_subscriptions: List[MutationSubscription] = []
        self.initial_assignment = True
        self.mutated: Set[str] = set()
        self.ingest(raw_state)

    def __repr__(self) -> str:
        return self.state.__repr__()

    def __contains__(self, key: str) -> bool:
        return self.state.__contains__(key)

    def ingest(self, raw_state: Dict) -> None:
        for key, raw_value in raw_state.items():
            self.__setitem__(key, raw_value)

    def items(self) -> Sequence[Tuple[str, Any]]:
        return cast(Sequence[Tuple[str, Any]], self.state.items())

    def get(self, key: str) -> Any:
        return self.state.get(key)

    def __getitem__(self, key: str) -> Any:
        return self.state.get(key)

    def __setitem__(self, key: str, raw_value: Any) -> None:
        with state_recursion_new(key):
            if not isinstance(key, str):
                raise ValueError(f"State keys must be strings. Received {str(key)} ({type(key)}).")
            old_value = self.state.get(key)
            self.state[key] = raw_value

            for local_mutation in self.local_mutation_subscriptions:
                if local_mutation.local_path == key:
                    if local_mutation.type == "subscription":
                        context_data = {"event": "mutation", "mutation": local_mutation.path}
                        payload = {"previous_value": old_value, "new_value": raw_value}

                        EventHandlerExecutor.invoke(
                            local_mutation.handler,
                            {
                                "state": local_mutation.state,
                                "context": context_data,
                                "payload": payload,
                                "session": _event_handler_session_info(),
                                "ui": _event_handler_ui_manager(),
                            },
                        )
                    elif local_mutation.type == "property":
                        assert local_mutation.property_name is not None
                        self[local_mutation.property_name] = local_mutation.handler(
                            local_mutation.state
                        )

            self._apply_raw(f"+{key}")

    def __delitem__(self, key: str) -> None:
        if key in self.state:
            del self.state[key]
            self._apply_raw(f"-{key}")  # Using "-" prefix to indicate deletion

    def remove(self, key: str) -> None:
        return self.__delitem__(key)

    def _apply_raw(self, key: str) -> None:
        self.mutated.add(key)

    def apply_mutation_marker(self, key: Optional[str] = None, recursive: bool = False) -> None:
        """
        Adds the mutation marker to a state. The mutation marker is used to track changes in the state.

        >>> self.apply_mutation_marker()

        Add the mutation marker on a specific field

        >>> self.apply_mutation_marker("field")

        Add the mutation marker to a state and all of its children

        >>> self.apply_mutation_marker(recursive=True)
        """
        keys = [key] if key is not None else self.state.keys()

        for k in keys:
            self._apply_raw(f"+{k}")
            if recursive is True:
                value = self.state[k]
                if isinstance(value, StateProxy):
                    value.apply_mutation_marker(recursive=True)

    @staticmethod
    def escape_key(key):
        return key.replace(".", r"\.")

    def get_mutations_as_dict(self) -> Dict[str, Any]:
        serialised_mutations: Dict[str, Union[Dict, List, str, bool, int, float, None]] = {}

        def carry_mutation_flag(base_key, child_key):
            child_mutation_flag, child_key = child_key[0], child_key[1:]
            return f"{child_mutation_flag}{base_key}.{child_key}"

        for key, value in list(self.state.items()):
            if key.startswith("_"):
                continue

            escaped_key = self.escape_key(key)
            serialised_value = None

            if isinstance(value, StateProxy):
                if f"+{key}" in self.mutated:
                    serialised_mutations[f"+{escaped_key}"] = serialised_value
                value.initial_assignment = False
                child_mutations = value.get_mutations_as_dict()
                if child_mutations is None:
                    continue
                for child_key, child_mutation in child_mutations.items():
                    nested_key = carry_mutation_flag(escaped_key, child_key)
                    serialised_mutations[nested_key] = child_mutation
            elif f"+{key}" in self.mutated:
                try:
                    serialised_value = state_serialiser.serialise(value)
                except BaseException:
                    raise ValueError(
                        f"""Couldn't serialise value of type "{ type(value) }" for key "{ key }"."""
                    )
                serialised_mutations[f"+{escaped_key}"] = serialised_value
            elif isinstance(value, MutableValue) is True and value.mutated():
                try:
                    serialised_value = state_serialiser.serialise(value)
                    value.reset_mutation()
                except BaseException:
                    raise ValueError(
                        f"""Couldn't serialise value of type "{ type(value) }" for key "{ key }"."""
                    )
                serialised_mutations[f"+{escaped_key}"] = serialised_value

        deleted_keys = {self.escape_key(key) for key in self.mutated if key.startswith("-")}
        for key in deleted_keys:
            serialised_mutations[f"{key}"] = None

        self.mutated = set()
        return serialised_mutations

    def to_dict(self) -> Dict[str, Any]:
        serialised = {}
        for key, value in self.state.items():
            if key.startswith("_"):
                continue
            serialised_value = None
            try:
                serialised_value = state_serialiser.serialise(value)
            except BaseException:
                raise ValueError(
                    f"""Couldn't serialise value of type "{ type(value) }" for key "{ key }"."""
                )
            serialised[key] = serialised_value
        return serialised

    def to_raw_state(self):
        """
        Converts a StateProxy and its children into a python dictionary.

        >>> state = State({'a': 1, 'c': {'a': 1, 'b': 3}})
        >>> _raw_state = state._state_proxy.to_raw_state()
        >>> {'a': 1, 'c': {'a': 1, 'b': 3}}

        :return: a python dictionary that represents the raw state
        """
        raw_state = {}
        for key, value in self.state.items():
            if isinstance(value, StateProxy):
                value = value.to_raw_state()
            raw_state[key] = value

        return raw_state


def get_annotations(instance) -> Dict[str, Any]:
    """
    Returns the annotations of the class in a way that works on python 3.9 and python 3.10
    """
    if isinstance(instance, type):
        ann = instance.__dict__.get("__annotations__", None)
    else:
        ann = getattr(instance, "__annotations__", None)

    if ann is None:
        ann = {}
    return ann


class StateMeta(type):
    """
    Constructs a class at runtime that extends WriterState or State
    with dynamic properties for each annotation of the class.
    """

    def __new__(cls, name, bases, attrs):
        klass = super().__new__(cls, name, bases, attrs)
        cls.bind_annotations_to_state_proxy(klass)
        return klass

    @classmethod
    def bind_annotations_to_state_proxy(cls, klass):
        """
        Loops through the class annotations and creates properties dynamically for each one.

        >>> class MyState(State):
        >>>     counter: int

        will be transformed into

        >>> class MyState(State):
        >>>
        >>>     @property
        >>>     def counter(self):
        >>>         return self._state_proxy["counter"]
        >>>
        >>>    @counter.setter
        >>>    def counter(self, value):
        >>>        self._state_proxy["counter"] = value

        Annotations that reference a State are ignored. The link will be established through a State instance
        when ingesting state data.

        >>> class MyAppState(State):
        >>>     title: str

        >>> class MyState(State):
        >>>     myapp: MyAppState # Nothing happens
        """

        annotations = get_annotations(klass)
        for key, expected_type in annotations.items():
            if key == "_state_proxy":
                raise AttributeError(
                    "_state_proxy is an reserved keyword for Writer Framework, don't use it in annotation."
                )

            if not (inspect.isclass(expected_type) and issubclass(expected_type, State)):
                proxy = DictPropertyProxy("_state_proxy", key)
                setattr(klass, key, proxy)


class State(metaclass=StateMeta):
    def __init__(self, raw_state: Optional[Dict[str, Any]] = None):
        final_raw_state = raw_state if raw_state is not None else {}

        self._state_proxy: StateProxy = StateProxy(final_raw_state)
        self.ingest(final_raw_state)

        # This step saves the properties associated with the instance
        for attribute in calculated_properties_per_state_type.get(self.__class__, []):
            getattr(self, attribute)

    def ingest(self, raw_state: Dict[str, Any]) -> None:
        """
        hydrates a state from raw data by applying a schema when it is provided.
        The existing content in the state is erased.


        >>> state = WriterState({'message': "hello world"})
        >>> state.ingest({'a': 1, 'b': 2})
        >>> {'a': 1, 'b': 2}
        """
        self._state_proxy.state = {}
        for key, value in raw_state.items():
            assert not isinstance(
                value, StateProxy
            ), f"state proxy datatype is not expected in ingest operation, {locals()}"
            self._set_state_item(key, value)

    def to_dict(self) -> dict:
        """
        Serializes state data as a dictionary

        Private attributes, prefixed with _, are ignored.

        >>> state = WriterState({'message': "hello world"})
        >>> return state.to_dict()
        """
        return self._state_proxy.to_dict()

    def to_raw_state(self) -> dict:
        """
        Converts a StateProxy and its children into a python dictionary that can be used to recreate the
        state from scratch.

        >>> state = WriterState({'a': 1, 'c': {'a': 1, 'b': 3}})
        >>> raw_state = state.to_raw_state()
        >>> "{'a': 1, 'c': {'a': 1, 'b': 3}}"

        :return: a python dictionary that represents the raw state
        """
        return self._state_proxy.to_raw_state()

    def __repr__(self) -> str:
        return self._state_proxy.__repr__()

    def __getitem__(self, key: str) -> Any:
        # Essential to support operation like
        # state['item']['a'] = state['item']['b']
        if hasattr(self, key):
            value = getattr(self, key)
            if isinstance(value, State):
                return value

        return self._state_proxy.__getitem__(key)

    def __setitem__(self, key: str, raw_value: Any) -> None:
        assert not isinstance(
            raw_value, StateProxy
        ), f"state proxy datatype is not expected, {locals()}"

        self._set_state_item(key, raw_value)

    def __delitem__(self, key: str) -> Any:
        return self._state_proxy.__delitem__(key)

    def remove(self, key: str) -> Any:
        return self.__delitem__(key)

    def items(self) -> Generator[Tuple[str, Any], None, None]:
        for k, v in self._state_proxy.items():
            if isinstance(v, StateProxy):
                # We don't want to expose StateProxy to the user, so
                # we replace it with relative State
                yield k, getattr(self, k)
            else:
                yield k, v

    def __contains__(self, key: str) -> bool:
        return self._state_proxy.__contains__(key)

    def _set_state_item(self, key: str, value: Any):
        """ """

        """
        At this level, the values that arrive are either States which encapsulate a StateProxy, or another datatype. 
        If there is a StateProxy, it is a fault in the code.
        """
        annotations = get_annotations(self)
        expected_type = cast(Type, annotations.get(key, None))
        expect_dict = _type_match_dict(expected_type)
        if isinstance(value, dict) and not expect_dict:
            """
            When the value is a dictionary and the attribute does not explicitly 
            expect a dictionary, we instantiate a new state to manage mutations.
            """
            state = annotations[key](value) if key in annotations else State()
            if not isinstance(state, State):
                raise ValueError(
                    f"Attribute {key} must inherit of State or requires a dict to accept dictionary"
                )

            setattr(self, key, state)
            state.ingest(value)
            self._state_proxy[key] = state._state_proxy
        else:
            if isinstance(value, State):
                value._state_proxy.apply_mutation_marker(recursive=True)
                self._state_proxy[key] = value._state_proxy
            else:
                self._state_proxy[key] = value

    def subscribe_mutation(
        self,
        path: Union[str, List[str]],
        handler: Callable[..., Union[None, Awaitable[None]]],
        initial_triggered: bool = False,
    ) -> None:
        r"""
        Automatically triggers a handler when a mutation occurs in the state.

        >>> def _increment_counter(state):
        >>>     state['my_counter'] += 1
        >>>
        >>> state = WriterState({'a': 1, 'c': {'a': 1, 'b': 3}, 'my_counter': 0})
        >>> state.subscribe_mutation('a', _increment_counter)
        >>> state.subscribe_mutation('c.a', _increment_counter)
        >>> state['a'] = 2 # will trigger _increment_counter
        >>> state['a'] = 3 # will trigger _increment_counter
        >>> state['c']['a'] = 2 # will trigger _increment_counter

        subscribe mutation accepts the signature of an event handler.

        >>> def _increment_counter(state, payload, context, session, ui):
        >>>     state['my_counter'] += 1
        >>>
        >>> state = WriterState({'a': 1, 'my_counter': 0})
        >>> state.subscribe_mutation('a', _increment_counter)
        >>> state['a'] = 2 # will trigger _increment_counter

        subscribe mutation accepts escaped dot expressions to encode key that contains `dot` separator

        >>> def _increment_counter(state, payload, context, session, ui):
        >>>     state['my_counter'] += 1
        >>>
        >>> state = WriterState({'a.b': 1, 'my_counter': 0})
        >>> state.subscribe_mutation('a\.b', _increment_counter)
        >>> state['a.b'] = 2 # will trigger _increment_counter

        :param path: path of mutation to monitor
        :param func: handler to call when the path is mutated
        """
        if isinstance(path, str):
            path_list = [path]
        else:
            path_list = path

        for p in path_list:
            state_proxy = self._state_proxy
            path_parts = parse_state_variable_expression(p)
            for i, path_part in enumerate(path_parts):
                if i == len(path_parts) - 1:
                    local_mutation = MutationSubscription("subscription", p, handler, self)
                    state_proxy.local_mutation_subscriptions.append(local_mutation)

                    # At startup, the application must be informed of the
                    # existing states. To cause this, we trigger manually
                    # the handler.
                    if initial_triggered is True:
                        EventHandlerExecutor.invoke(
                            handler,
                            {
                                "state": self,
                                "context": {"event": "init"},
                                "payload": {},
                                "session": {},
                                "ui": _event_handler_ui_manager(),
                            },
                        )

                elif path_part in state_proxy:
                    state_proxy = state_proxy[path_part]
                else:
                    raise ValueError(f"Mutation subscription failed - {p} not found in state")

    def calculated_property(
        self,
        property_name: str,
        path: Union[str, List[str]],
        handler: Callable[..., Union[None, Awaitable[None]]],
    ) -> None:
        """
        Update a calculated property when a mutation triggers

        This method is dedicated to be used through a calculated property. It is not
        recommended to invoke it directly.

        >>> class MyState(State):
        >>>     title: str
        >>>
        >>>     wf.property('title')
        >>>     def title_upper(self):
        >>>         return self.title.upper()

        Usage
        =====

        >>> state = wf.init_state({'title': 'hello world'})
        >>> state.calculated_property('title_upper', 'title', lambda state: state.title.upper())
        """
        if isinstance(path, str):
            path_list = [path]
        else:
            path_list = path

        for p in path_list:
            state_proxy = self._state_proxy
            path_parts = parse_state_variable_expression(p)
            for i, path_part in enumerate(path_parts):
                if i == len(path_parts) - 1:
                    local_mutation = MutationSubscription(
                        "property", p, handler, self, property_name
                    )
                    state_proxy.local_mutation_subscriptions.append(local_mutation)
                    state_proxy[property_name] = handler(self)
                elif path_part in state_proxy:
                    state_proxy = state_proxy[path_part]
                else:
                    raise ValueError(f"Property subscription failed - {p} not found in state")


class WriterState(State):
    """
    Root state. Comprises user configurable state and
    mail (notifications, log entries, etc).
    """

    LOG_ENTRY_MAX_LEN = 8192

    def __init__(self, raw_state: Dict[str, Any] = {}, mail: List[Any] = []):
        super().__init__(raw_state)
        self.mail = copy.deepcopy(mail)

    @property
    def user_state(self) -> StateProxy:
        return self._state_proxy

    @classmethod
    def get_new(cls):
        """Returns a new WriterState instance set to the initial state."""

        return initial_state.get_clone()

    def get_clone(self) -> "WriterState":
        """
        get_clone clones the destination application state for the session.

        The class is rebuilt identically in the case where the user
        has constructed a schema inherited from WriterState

        >>> class AppSchema(WriterState):
        >>>     counter: int
        >>>
        >>> root_state = AppSchema({'counter': 1})
        >>> clone_state = root_state.get_clone() # instance of AppSchema
        """
        try:
            cloned_user_state = copy.deepcopy(self.user_state.to_raw_state())
            cloned_mail = copy.deepcopy(self.mail)
        except BaseException:
            substitute_state = WriterState()
            substitute_state.add_log_entry(
                "error",
                "Cannot clone state",
                "The state may contain unpickable objects, such as modules.",
                traceback.format_exc(),
            )
            return substitute_state

        cloned_state = self.__class__(cloned_user_state, cloned_mail)
        _clone_mutation_subscriptions(cloned_state, self)
        return cloned_state

    def add_mail(self, type: str, payload: Any) -> None:
        mail_item = {"type": type, "payload": payload}
        self.mail.append(mail_item)

    def add_notification(
        self, type: Literal["info", "success", "warning", "error"], title: str, message: str
    ) -> None:
        self.add_mail(
            "notification",
            {
                "type": type,
                "title": title,
                "message": message,
            },
        )

    def _log_entry_in_logger(
        self,
        type: Literal["debug", "info", "warning", "error", "critical"],
        title: str,
        message: str,
        code: Optional[str] = None,
        blueprint_execution: Optional[BlueprintExecutionLog] = None,
    ) -> None:
        if not Config.logger:
            return
        log_args: Tuple[str, ...] = ()

        if code:
            log_args = (title, message, code)
        else:
            log_args = (title, message)

        log_colors = {
            "debug": "\x1b[36;20m",  # Cyan for debug
            "info": "\x1b[34;20m",  # Blue for info
            "warning": "\x1b[33;20m",  # Yellow for warning
            "error": "\x1b[31;20m",  # Red for error
            "critical": "\x1b[35;20m",  # Magenta for critical
        }

        log_methods = {
            "debug": Config.logger.debug,
            "info": Config.logger.info,
            "warning": Config.logger.warning,
            "error": Config.logger.error,
            "critical": Config.logger.critical,
        }

        log_message = "From app log: " + ("\n%s" * len(log_args))

        if blueprint_execution:
            log_message += "\n"
            for entry in blueprint_execution.summary:
                outcome = entry.get("outcome")
                if outcome is None:
                    continue
                component_id = entry.get("componentId")
                entry_message = entry.get("message")
                log_message += f"- Id: {component_id} | Outcome: {outcome} | {entry_message}\n"

        color = log_colors.get(type, "\x1b[0m")  # Default to no color if type not found
        log_method = log_methods.get(
            type, Config.logger.info
        )  # Default to info level if type not found

        log_method(f"{color}{log_message}\x1b[0m", *log_args)

    def add_log_entry(
        self,
        type: Literal["info", "error"],
        title: str,
        message: str,
        code: Optional[str] = None,
        blueprint_execution: Optional[BlueprintExecutionLog] = None,
        id: Optional[str] = None,
    ) -> None:
        self._log_entry_in_logger(type, title, message, code, blueprint_execution)
        if not Config.is_mail_enabled_for_log:
            return
        shortened_message = None
        if len(message) > WriterState.LOG_ENTRY_MAX_LEN:
            shortened_message = message[0 : WriterState.LOG_ENTRY_MAX_LEN] + "..."
        else:
            shortened_message = message

        if id is not None:
            self._remove_duplicates_in_mail(id)

        self.add_mail(
            "logEntry",
            {
                "type": type,
                "title": title,
                "message": shortened_message,
                "code": code,
                "blueprintExecution": blueprint_execution,
                "id": id,
            },
        )

    def file_download(self, data: Any, file_name: str):
        if not isinstance(data, (bytes, FileWrapper, BytesWrapper)):
            raise ValueError(
                "Data for a fileDownload mail must be bytes, a FileWrapper or a BytesWrapper."
            )
        self.add_mail(
            "fileDownload", {"data": state_serialiser.serialise(data), "fileName": file_name}
        )

    def open_url(self, url: str):
        self.add_mail("openUrl", url)

    def clear_mail(self) -> None:
        self.mail = []

    def _remove_duplicates_in_mail(self, id: str) -> None:
        new_mail = []
        for entry in self.mail:
            if entry["type"] != "logEntry":
                new_mail.append(entry)
                continue
            log_id = entry["payload"].get("id")
            if log_id != id:
                new_mail.append(entry)
                continue
        self.mail = new_mail

    def set_page(self, active_page_key: str) -> None:
        self.add_mail("pageChange", active_page_key)

    def set_route_vars(self, route_vars: Dict[str, str]) -> None:
        self.add_mail("routeVarsChange", route_vars)

    def import_stylesheet(self, stylesheet_key: str, path: str) -> None:
        self.add_mail("importStylesheet", {"stylesheetKey": stylesheet_key, "path": path})

    def import_script(self, script_key: str, path: str) -> None:
        """
        imports the content of a script into the page

        >>> initial_state = wf.init_state({
        >>>     "counter": 1
        >>> })
        >>>
        >>> initial_state.import_script("my_script", "/static/script.js")
        """
        self.add_mail("importScript", {"scriptKey": script_key, "path": path})

    def import_frontend_module(self, module_key: str, specifier: str) -> None:
        self.add_mail("importModule", {"moduleKey": module_key, "specifier": specifier})

    def call_frontend_function(self, module_key: str, function_name: str, args: List) -> None:
        self.add_mail(
            "functionCall", {"moduleKey": module_key, "functionName": function_name, "args": args}
        )


class MiddlewareExecutor:
    """
    A MiddlewareExecutor executes middleware in a controlled context. It allows writer framework
    to manage different implementations of middleware.

    Case 1 : A middleware is a generator, then run before and after code

    >>> @wf.middleware()
    >>> def my_middleware():
    >>>     print("before event handler")
    >>>     yield()
    >>>     print("after event handler")

    Case 2 : A middleware is just a function, then run the function before

    >>> @wf.middleware()
    >>> def my_middleware():
    >>>     print("before event handler")
    """

    def __init__(self, middleware: Callable):
        self.middleware = middleware

    @contextlib.contextmanager
    def execute(self, args: dict):
        middleware_args = EventHandlerExecutor.build_arguments(self.middleware, args)
        it = self.middleware(*middleware_args)
        try:
            yield from it
        except StopIteration:
            yield
        except TypeError:
            yield


class MiddlewareRegistry:
    def __init__(self) -> None:
        self.registry: List[MiddlewareExecutor] = []

    def register(self, middleware: Callable):
        me = MiddlewareExecutor(middleware)
        self.registry.append(me)

    def executors(self) -> List[MiddlewareExecutor]:
        """
        Retrieves middlewares prepared for execution

        >>> executors = middleware_registry.executors()
        >>> result = writer_event_handler_invoke_with_middlewares(executors, lambda state: pass, {'state': {}, 'payload': {}})
        """
        return self.registry


class EventHandlerRegistry:
    """
    Maps functions registered as event handlers from the user app's core
    and external modules, providing an access mechanism to these maps.
    """

    class HandlerMeta(TypedDict):
        name: str
        args: List[str]

    class HandlerEntry(TypedDict):
        callable: Callable
        meta: "EventHandlerRegistry.HandlerMeta"

    # === BLUEPRINT HANLDERS ===
    @staticmethod
    def stop_blueprint_run(payload: dict, blueprint_runner: "BlueprintRunner"):
        run_id = payload.pop("run_id", None)
        if not run_id:
            raise ValueError("Missing run_id in payload")
        return blueprint_runner.cancel_blueprint_execution(run_id=run_id)

    @staticmethod
    def run_blueprint_by_id(
        payload: dict,
        context: dict,
        session: dict,
        blueprint_runner: "BlueprintRunner",
        vault: Dict,
    ):
        blueprint_id = payload.pop("blueprint_id", None)
        if not blueprint_id:
            raise ValueError("Missing blueprint_id in payload")
        execution_environment = EventHandler._get_blueprint_execution_environment(
            payload, context, session, vault
        )
        return blueprint_runner.run_blueprint(
            component_id=blueprint_id,
            execution_environment=execution_environment,
            title="Blueprint execution triggered on demand",
        )

    @staticmethod
    def run_blueprint_by_key(
        payload: dict,
        context: dict,
        session: dict,
        blueprint_runner: "BlueprintRunner",
        vault: Dict,
    ):
        blueprint_key = payload.pop("blueprint_key", None)
        if not blueprint_key:
            raise ValueError("Missing blueprint_key in payload")
        execution_environment = EventHandler._get_blueprint_execution_environment(
            payload, context, session, vault
        )
        return blueprint_runner.run_blueprint_by_key(
            blueprint_key=blueprint_key, execution_environment=execution_environment
        )

    @staticmethod
    def run_blueprint_via_api(
        payload: dict,
        context: dict,
        session: dict,
        blueprint_runner: "BlueprintRunner",
        vault: Dict,
    ):
        """
        This handler is used to run a blueprint via the API.
        It is used by the frontend to run a blueprint when the user clicks on a button.
        """
        blueprint_id = payload.pop("blueprint_id", None)
        if not blueprint_id:
            raise ValueError("Missing blueprint_id in payload")
        execution_environment = EventHandler._get_blueprint_execution_environment(
            payload, context, session, vault
        )
        return blueprint_runner.run_blueprint_via_api(
            blueprint_id=blueprint_id,
            branch_id=payload.pop("branch_id", None),
            execution_environment=execution_environment,
        )

    @staticmethod
    def run_blueprint_branch(
        payload: dict,
        context: dict,
        session: dict,
        blueprint_runner: "BlueprintRunner",
        vault: Dict,
    ):
        branch_id = payload.pop("branch_id", None)
        if not branch_id:
            raise ValueError("Missing branch_id in payload")
        execution_environment = EventHandler._get_blueprint_execution_environment(
            payload, context, session, vault
        )
        return blueprint_runner.run_branch(
            start_node_id=branch_id,
            branch_out_id=None,
            execution_environment=execution_environment,
            title="Branch execution triggered by demand",
        )

    def __init__(self):
        self.handler_map: Dict[str, "EventHandlerRegistry.HandlerEntry"] = {
            "stop_blueprint_run": {
                "callable": self.stop_blueprint_run,
                "meta": {"name": "stop_blueprint_run", "args": ["payload", "blueprint_runner"]},
            },
            "run_blueprint_by_key": {
                "callable": self.run_blueprint_by_key,
                "meta": {
                    "name": "run_blueprint_by_key",
                    "args": ["payload", "context", "session", "blueprint_runner", "vault"],
                },
            },
            "run_blueprint_by_id": {
                "callable": self.run_blueprint_by_id,
                "meta": {
                    "name": "run_blueprint_by_id",
                    "args": ["payload", "context", "session", "blueprint_runner", "vault"],
                },
            },
            "run_blueprint_via_api": {
                "callable": self.run_blueprint_via_api,
                "meta": {
                    "name": "run_blueprint_via_api",
                    "args": ["payload", "context", "session", "blueprint_runner", "vault"],
                },
            },
            "run_blueprint_branch": {
                "callable": self.run_blueprint_branch,
                "meta": {
                    "name": "run_blueprint_branch",
                    "args": ["payload", "context", "session", "blueprint_runner", "vault"],
                },
            },
        }

    def __iter__(self):
        return iter(self.handler_map.keys())

    def register_handler(self, handler: Callable):
        module_name = handler.__module__

        # Prepare "access name"
        # (i.e. the key that frontend uses to retrieve handler)
        if module_name == "writeruserapp":
            # Use the handler's __qualname__ directly
            # for functions from main.py in user's app
            access_name = handler.__qualname__
        else:
            # For external handlers, separate the module name
            # and handler __qualname__by a dot
            access_name = f"{module_name}.{handler.__qualname__}"

        entry: EventHandlerRegistry.HandlerEntry = {
            "callable": handler,
            "meta": {"name": access_name, "args": inspect.getfullargspec(handler).args},
        }

        self.handler_map[access_name] = entry

    def register_module(self, module: ModuleType):
        if isinstance(module, ModuleType):
            all_fn_names = (x[0] for x in inspect.getmembers(module, inspect.isfunction))
            exposed_fn_names = list(filter(lambda x: not x.startswith("_"), all_fn_names))

            for fn_name in exposed_fn_names:
                fn_callable = getattr(module, fn_name)
                if not fn_callable:
                    continue
                self.register_handler(fn_callable)
        else:
            raise ValueError(f"Attempted to register a non-module object: {module}")

    def find_handler_callable(self, handler_name: str) -> Optional[Callable]:
        if handler_name not in self.handler_map:
            return None
        handler_entry: EventHandlerRegistry.HandlerEntry = self.handler_map[handler_name]
        return handler_entry["callable"]

    def get_handler_meta(self, handler_name: str) -> "EventHandlerRegistry.HandlerMeta":
        if handler_name not in self.handler_map:
            raise RuntimeError(f"Handler {handler_name} is not registered")
        entry: EventHandlerRegistry.HandlerEntry = self.handler_map[handler_name]
        return entry["meta"]

    def gather_handler_meta(self) -> List["EventHandlerRegistry.HandlerMeta"]:
        return [self.get_handler_meta(handler_name) for handler_name in self]


class EventDeserialiser:
    """Applies transformations to the payload of an incoming event, depending on its type.

    The transformation happens in place: the event passed to the transform method is mutated.

    Its main goal is to deserialise incoming content in a controlled and predictable way,
    applying sanitisation of inputs where relevant."""

    def __init__(self, session: "WriterSession"):
        self.evaluator = writer.evaluator.Evaluator(
            session.session_state, session.session_component_tree
        )

    def transform(self, ev: WriterEvent) -> None:
        # Events without payloads are safe
        # This includes non-custom events such as click
        # Events not natively provided by Writer Framework aren't sanitised

        if ev.payload is None or not ev.type.startswith("wf-"):
            return

        # Look for a method in this class that matches the event type
        # As a security measure, all event types starting with "wf-" must be linked to a transformer function.

        custom_event_name = ev.type[3:]
        func_name = "_transform_" + custom_event_name.replace("-", "_")
        if not hasattr(self, func_name):
            if ev.isSafe:
                return
            ev.payload = {}
            raise ValueError("No payload transformer available for custom event type.")
        tf_func = getattr(self, func_name)
        try:
            tf_payload = tf_func(ev)
        except BaseException:
            ev.payload = {}
            raise RuntimeError("Payload transformation failed.")
        else:
            ev.payload = tf_payload

    def _transform_tag_click(self, ev: WriterEvent) -> Optional[str]:
        payload = ev.payload
        instance_path = ev.instancePath
        if not instance_path:
            raise ValueError("This event cannot be run as a global event.")
        options = self.evaluator.evaluate_field(instance_path, "tags", True, "{ }")
        if not isinstance(options, dict):
            raise ValueError("Invalid value for tags")
        if payload not in options.keys():
            raise ValueError("Unauthorised option")
        return payload

    def _transform_option_change(self, ev: WriterEvent) -> Optional[str]:
        payload = ev.payload
        instance_path = ev.instancePath
        if not instance_path:
            raise ValueError("This event cannot be run as a global event.")
        options = self.evaluator.evaluate_field(
            instance_path, "options", True, """{ "a": "Option A", "b": "Option B" }"""
        )
        if not isinstance(options, dict):
            raise ValueError("Invalid value for options")
        if payload not in options.keys():
            raise ValueError("Unauthorised option")
        return payload

    def _transform_options_change(self, ev: WriterEvent) -> Optional[List[str]]:
        payload = ev.payload
        instance_path = ev.instancePath
        if not instance_path:
            raise ValueError("This event cannot be run as a global event.")
        options = self.evaluator.evaluate_field(
            instance_path, "options", True, """{ "a": "Option A", "b": "Option B" }"""
        )
        if not isinstance(options, dict):
            raise ValueError("Invalid value for options")
        if not isinstance(payload, list):
            raise ValueError("Invalid multiple options payload. Expected a list.")
        if not all(item in options.keys() for item in payload):
            raise ValueError("Unauthorised option")
        return payload

    def _transform_toggle(self, ev: WriterEvent) -> bool:
        payload = bool(ev.payload)
        return payload

    def _transform_keydown(self, ev) -> Dict:
        payload = ev.payload
        key = str(payload.get("key"))
        ctrl_key = bool(payload.get("ctrlKey"))
        shift_key = bool(payload.get("shiftKey"))
        meta_key = bool(payload.get("metaKey"))
        tf_payload = {
            "key": key,
            "ctrl_key": ctrl_key,
            "shift_key": shift_key,
            "meta_key": meta_key,
        }
        return tf_payload

    def _transform_click(self, ev) -> Dict:
        payload = ev.payload
        ctrl_key = bool(payload.get("ctrlKey"))
        shift_key = bool(payload.get("shiftKey"))
        meta_key = bool(payload.get("metaKey"))
        tf_payload = {"ctrl_key": ctrl_key, "shift_key": shift_key, "meta_key": meta_key}
        return tf_payload

    def _transform_hashchange(self, ev) -> Dict:
        payload = ev.payload
        page_key = payload.get("pageKey")
        route_vars = dict(payload.get("routeVars"))
        tf_payload = {"page_key": page_key, "route_vars": route_vars}
        return tf_payload

    def _transform_page_open(self, ev) -> str:
        payload = str(ev.payload)
        return payload

    def _transform_app_open(self, ev) -> dict:
        payload = ev.payload
        page_key = payload.get("pageKey")
        route_vars = dict(payload.get("routeVars"))
        tf_payload = {"page_key": page_key, "route_vars": route_vars}
        return tf_payload

    def _transform_chatbot_message(self, ev) -> dict:
        payload = dict(ev.payload)
        return payload

    def _transform_chatbot_action_click(self, ev) -> str:
        payload = str(ev.payload)
        return payload

    def _transform_change(self, ev) -> str:
        payload = str(ev.payload)
        return payload

    def _transform_change_finish(self, ev) -> str:
        return self._transform_change(ev)

    def _transform_number_change(self, ev) -> Optional[float]:
        try:
            return float(ev.payload)
        except ValueError:
            return None

    def _transform_number_change_finish(self, ev) -> Optional[float]:
        return self._transform_number_change(ev)

    def _transform_webcam(self, ev) -> Any:
        return urllib.request.urlopen(ev.payload).read()

    def _file_item_transform(self, file_item: WriterFileItem) -> Dict:
        data = file_item.get("data")
        if data is None:
            raise ValueError("No data provided.")
        return {
            "name": file_item.get("name"),
            "type": file_item.get("type"),
            "data": urllib.request.urlopen(data).read(),
        }

    def _transform_file_change(self, ev) -> List[Dict]:
        payload = ev.payload
        tf_payload = list(map(self._file_item_transform, payload))

        return tf_payload

    def _transform_date_change(self, ev) -> str:
        payload = ev.payload

        if not isinstance(payload, str):
            raise ValueError("Date must be a string.")
        try:
            datetime.date.fromisoformat(payload)
        except ValueError:
            raise ValueError("Date must be in YYYY-MM-DD format or another valid ISO 8601 format.")

        return payload

    def _transform_time_change(self, ev) -> str:
        payload = ev.payload

        if not isinstance(payload, str):
            raise ValueError("Time must be a string.")
        try:
            time.strptime(payload, "%H:%M")
        except ValueError:
            raise ValueError(
                "Time must be in hh:mm format (in 24-hour format that includes leading zeros)."
            )

        return payload

    def _transform_range_change(self, ev) -> list[int]:
        payload = ev.payload

        if not isinstance(payload, list):
            raise ValueError("Range must be an array.")

        if len(payload) != 2:
            raise ValueError("Range must contains exactly two values.")

        if not isinstance(payload[0], numbers.Real):
            raise ValueError("First item is not a number.")

        if not isinstance(payload[1], numbers.Real):
            raise ValueError("Second item is not a number.")

        if payload[0] > payload[1]:
            raise ValueError("First item is higher than second.")

        return payload

    def _transform_change_page_size(self, ev) -> Optional[int]:
        try:
            return int(ev.payload)
        except ValueError:
            return None

    def _transform_change_page(self, ev) -> Optional[int]:
        try:
            return int(ev.payload)
        except ValueError:
            return None

    def _transform_dataframe_update(self, ev: WriterEvent) -> Optional[Dict]:
        payload = ev.payload
        if not isinstance(payload, dict):
            return None

        payload = _deserialize_bigint_format(payload)
        return payload

    def _transform_dataframe_add(self, ev: WriterEvent) -> Optional[Dict]:
        payload = ev.payload
        if not isinstance(payload, dict):
            return None

        payload = _deserialize_bigint_format(payload)
        return payload

    def _transform_dataframe_action(self, ev: WriterEvent) -> Optional[Dict]:
        payload = ev.payload
        if not isinstance(payload, dict):
            return None

        payload = _deserialize_bigint_format(payload)
        return payload

    def _transform_tab_change(self, ev: WriterEvent) -> Optional[str]:
        payload = ev.payload
        if not isinstance(payload, str):
            return None
        return payload


class SessionManager:
    """
    Stores and manages sessions.
    """

    IDLE_SESSION_MAX_SECONDS = 3600
    TOKEN_SIZE_BYTES = 32
    hex_pattern = re.compile(r"^[0-9a-fA-F]{" + str(TOKEN_SIZE_BYTES * 2) + r"}$")

    def __init__(self) -> None:
        self.sessions: Dict[str, WriterSession] = {}
        self.verifiers: List[Callable] = []

    def add_verifier(self, verifier: Callable) -> None:
        self.verifiers.append(verifier)

    def _verify_before_new_session(
        self, cookies: Optional[Dict] = None, headers: Optional[Dict] = None
    ) -> bool:
        for verifier in self.verifiers:
            args = inspect.getfullargspec(verifier).args
            arg_values = []
            for arg in args:
                if arg == "cookies":
                    arg_values.append(cookies)
                elif arg == "headers":
                    arg_values.append(headers)
            verifier_result = verifier(*arg_values)
            if verifier_result is False:
                return False
            elif verifier_result is True:
                pass
            else:
                raise ValueError("Invalid verifier return value. Must be True or False.")
        return True

    def _check_proposed_session_id(self, proposed_session_id: Optional[str]) -> bool:
        if proposed_session_id is None:
            return True
        if SessionManager.hex_pattern.match(proposed_session_id):
            return True
        return False

    def get_new_session(
        self,
        cookies: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        proposed_session_id: Optional[str] = None,
    ) -> Optional[WriterSession]:
        if not self._check_proposed_session_id(proposed_session_id):
            return None
        if not self._verify_before_new_session(cookies, headers):
            return None
        new_id = None
        if proposed_session_id is None:
            new_id = self._generate_session_id()
        else:
            new_id = proposed_session_id
        new_session = WriterSession(new_id, cookies, headers)
        self.sessions[new_id] = new_session
        return new_session

    def get_session(
        self, session_id: Optional[str], restore_initial_mail: bool = False
    ) -> Optional[WriterSession]:
        if session_id is None:
            return None
        if session_id == "anonymous":
            return None

        session = self.sessions.get(session_id)
        if session is not None and restore_initial_mail is True:
            session.session_state.mail = copy.copy(initial_state.mail)

        return session

    def _generate_session_id(self) -> str:
        return secrets.token_hex(SessionManager.TOKEN_SIZE_BYTES)

    def clear_all(self) -> None:
        self.sessions = {}

    def close_session(self, session_id: str) -> None:
        if session_id not in self.sessions:
            return
        del self.sessions[session_id]

    def prune_sessions(self) -> None:
        cutoff_timestamp = int(time.time()) - SessionManager.IDLE_SESSION_MAX_SECONDS
        prune_sessions = []
        for session_id, session in self.sessions.items():
            if session.last_active_timestamp < cutoff_timestamp:
                prune_sessions.append(session_id)
        for session_id in prune_sessions:
            self.close_session(session_id)

    @staticmethod
    def generate_session_id() -> str:
        """
        Generates a random session identifier which can be used to propose a session number before starting
        the app process in the apiinit route.
        """
        return secrets.token_hex(SessionManager.TOKEN_SIZE_BYTES)


class EventHandler:
    """
    Handles events in the context of a Session.
    """

    def __init__(self, session: WriterSession) -> None:
        import writer.blueprints

        self.session = session
        self.session_state = session.session_state
        self.session_component_tree = session.session_component_tree
        self.deser = EventDeserialiser(session)
        self.evaluator = writer.evaluator.Evaluator(
            session.session_state, session.session_component_tree
        )
        self.blueprint_runner = writer.blueprints.BlueprintRunner(session)

    def _handle_binding(self, event_type, target_component, instance_path, payload) -> None:
        if not target_component.binding:
            return
        binding = target_component.binding
        if binding["eventType"] != event_type:
            return
        self.evaluator.set_state(binding["stateRef"], instance_path, payload)

    @staticmethod
    def _get_blueprint_execution_environment(payload, context, session, vault):
        return {
            "payload": payload,
            "context": context,
            "session": session,
            "vault": vault,
        }

    def _get_blueprint_callable(
        self,
        blueprint_key: Optional[str] = None,
        blueprint_id: Optional[str] = None,
        branch_id: Optional[str] = None,
    ):
        def fn(payload, context, session, vault):
            execution_environment = self._get_blueprint_execution_environment(
                payload, context, session, vault
            )
            if blueprint_key:
                return self.blueprint_runner.run_blueprint_by_key(
                    blueprint_key, execution_environment
                )
            elif blueprint_id:
                return self.blueprint_runner.run_blueprint(
                    blueprint_id, execution_environment, "Blueprint execution triggered on demand"
                )
            elif branch_id:
                return self.blueprint_runner.run_branch(
                    branch_id,
                    None,
                    execution_environment,
                    "Blueprint branch execution triggered on demand",
                )

        return fn

    def _get_handler_callable(self, handler: str) -> Optional[Callable]:
        current_app_process = get_app_process()
        handler_registry = current_app_process.handler_registry
        callable_handler = handler_registry.find_handler_callable(handler)
        return callable_handler

    def _get_calling_arguments(self, ev: WriterEvent, instance_path: Optional[InstancePath] = None):
        context_data = self.evaluator.get_context_data(instance_path) if instance_path else {}
        context_data["event"] = ev.type
        return {
            "state": self.session_state,
            "payload": ev.payload,
            "context": context_data,
            "session": _event_handler_session_info(),
            "ui": _event_handler_ui_manager(),
            "blueprint_runner": self.blueprint_runner,
            "vault": writer_vault.get_secrets(),
        }

    def _call_handler_callable(self, handler_callable: Callable, calling_arguments: Dict) -> Any:
        current_app_process = get_app_process()
        result = None
        with (
            core_ui.use_component_tree(self.session.session_component_tree),
            use_stdout_redirect(
                lambda entry: self.session_state.add_log_entry("info", "Stdout message", entry)
            ),
        ):
            middlewares_executors = current_app_process.middleware_registry.executors()
            result = EventHandlerExecutor.invoke_with_middlewares(
                middlewares_executors, handler_callable, calling_arguments
            )

        return result

    def _deserialize(self, ev: WriterEvent):
        try:
            self.deser.transform(ev)
        except BaseException as e:
            self.session_state.add_notification(
                "error",
                "Error",
                f"A deserialization error occurred when handling event '{ ev.type }'.",
            )
            self.session_state.add_log_entry(
                "error",
                "Deserialization Failed",
                f"The data sent might be corrupt. A runtime exception was raised when deserializing event '{ ev.type }'.",
                traceback.format_exc(),
            )
            raise e

    def _handle_global_event(self, ev: WriterEvent):
        try:
            if not ev.isSafe:
                raise PermissionError("Attempted executing a global event in an unsafe context.")
            if not ev.handler:
                raise ValueError("Handler not specified when attempting to execute global event.")
            handler_callable = self._get_handler_callable(ev.handler)
            if not handler_callable:
                return
            calling_arguments = self._get_calling_arguments(ev, instance_path=None)
            return self._call_handler_callable(handler_callable, calling_arguments)
        except BaseException as e:
            if not isinstance(e, BlueprintExecutionError):
                # Only create a notification and log entry
                # for non-blueprint errors, as blueprint errors
                # have their own logging mechanism
                self.session_state.add_notification(
                    "error",
                    "Runtime Error",
                    f"An error occurred when processing event '{ ev.type }'.",
                )
                self.session_state.add_log_entry(
                    "error",
                    "Runtime Exception",
                    f"A runtime exception was raised when processing event '{ ev.type }'.",
                    traceback.format_exc(),
                )
            raise e

    def _handle_component_event(self, ev: WriterEvent):
        instance_path = ev.instancePath
        try:
            if not instance_path:
                raise ValueError("Component event must specify an instance path.")
            target_id = instance_path[-1]["componentId"]
            target_component = cast(Component, self.session_component_tree.get_component(target_id))
            self._handle_binding(ev.type, target_component, instance_path, ev.payload)
            calling_arguments = self._get_calling_arguments(ev, instance_path)
            execution_environment = self._get_blueprint_execution_environment(
                calling_arguments.get("payload"),
                calling_arguments.get("context"),
                calling_arguments.get("session"),
                calling_arguments.get("vault"),
            )
            self.blueprint_runner.execute_ui_trigger(target_id, ev.type, execution_environment)
            if not target_component.handlers:
                return None
            handler = target_component.handlers.get(ev.type)
            if not handler:
                return None
            handler_callable = self._get_handler_callable(handler)
            if not handler_callable:
                return
            return self._call_handler_callable(handler_callable, calling_arguments)
        except BaseException as e:
            self.session_state.add_notification(
                "error",
                "Runtime Error",
                f"An error occurred when processing event '{ ev.type }'.",
            )
            self.session_state.add_log_entry(
                "error",
                "Runtime Exception",
                f"A runtime exception was raised when processing event '{ ev.type }'.",
                traceback.format_exc(),
            )
            raise e

    def handle(self, ev: WriterEvent) -> WriterEventResult:
        try:
            if not ev.isSafe and ev.handler is not None:
                raise PermissionError("Unexpected handler set on event.")
            self._deserialize(ev)
            if not ev.instancePath:
                return {"ok": True, "result": self._handle_global_event(ev)}
            else:
                return {"ok": True, "result": self._handle_component_event(ev)}
        except BaseException as e:
            return {"ok": False, "result": str(e)}


class EventHandlerExecutor:
    @staticmethod
    def build_arguments(func: Callable, writer_args: dict) -> List[Any]:
        """
        Constructs the list of arguments based on the signature of the function
        which can be a handler or middleware.

        >>> def my_event_handler(state, context):
        >>>     yield

        >>> args = EventHandlerExecutor.build_arguments(my_event_handler, {'state': {}, 'payload': {}, 'context': {"target": '11'}, 'session': None, 'ui': None})
        >>> [{}, {"target": '11'}]

        :param func: the function that will be called
        :param writer_args: the possible arguments in writer (state, payload, ...)
        """
        handler_args = inspect.getfullargspec(func).args
        func_args = []
        for arg in handler_args:
            if arg in writer_args:
                func_args.append(writer_args[arg])

        return func_args

    @staticmethod
    def invoke(callable_handler: Callable, writer_args: dict) -> Any:
        """
        Runs a handler based on its signature.

        If the handler is asynchronous, it is executed asynchronously.
        If the handler only has certain parameters, only these are passed as arguments

        >>> def my_handler(state):
        >>>     state['a'] = 2
        >>>
        >>> EventHandlerExecutor.invoke(my_handler, {'state': {'a': 1}, 'payload': None, 'context': None, 'session': None, 'ui': None})
        """
        is_async_handler = inspect.iscoroutinefunction(callable_handler)
        if not callable(callable_handler) and not is_async_handler:
            raise ValueError("Invalid handler. The handler isn't a callable object.")

        handler_args = EventHandlerExecutor.build_arguments(callable_handler, writer_args)

        if is_async_handler:
            async_wrapper = _async_wrapper_internal(callable_handler, handler_args)
            result = asyncio.run(async_wrapper)
        else:
            result = callable_handler(*handler_args)

        return result

    @staticmethod
    def invoke_with_middlewares(
        middlewares_executors: List[MiddlewareExecutor],
        callable_handler: Callable,
        writer_args: dict,
    ) -> Any:
        """
        Runs the middlewares then the handler. This function allows you to manage exceptions that are triggered in middleware

        :param middlewares_executors: The list of middleware to run
        :param callable_handler: The target handler

        >>> @wf.middleware()
        >>> def my_middleware(state, payload, context, session, ui):
        >>>     yield

        >>> executor = MiddlewareExecutor(my_middleware, {'state': {}, 'payload': None, 'context': None, 'session': None, 'ui': None})
        >>> EventHandlerExecutor.invoke_with_middlewares([executor], my_handler, {'state': {}, 'payload': None, 'context': None, 'session': None, 'ui': None}
        """
        if len(middlewares_executors) == 0:
            return EventHandlerExecutor.invoke(callable_handler, writer_args)
        else:
            executor = middlewares_executors[0]
            with executor.execute(writer_args):
                return EventHandlerExecutor.invoke_with_middlewares(
                    middlewares_executors[1:], callable_handler, writer_args
                )


class DictPropertyProxy:
    """
    A descriptor based recipe that makes it possible to write shorthands
    that forward attribute access from one object onto another.

    >>> class A:
    >>>     foo: int = DictPropertyProxy("proxy_state", "prop1")
    >>>     bar: int = DictPropertyProxy("proxy_state", "prop2")
    >>>
    >>>     def __init__(self):
    >>>         self._state_proxy = StateProxy({"prop1": 1, "prop2": 2})
    >>>
    >>> a = A()
    >>> print(a.foo)

    This descriptor avoids writing the code below to establish a proxy
     with a child instance

    >>> class A:
    >>>
    >>>     def __init__(self):
    >>>         self._state_proxy = StateProxy({"prop1": 1, "prop2": 2})
    >>>
    >>>     @property
    >>>     def prop1(self):
    >>>         return self._state_proxy['prop1']
    >>>
    >>>     @foo.setter
    >>>     def prop1(self, value):
    >>>         self._state_proxy['prop1'] = value
    >>>
    """

    def __init__(self, objectName, key):
        self.objectName = objectName
        self.key = key

    def __get__(self, instance, owner=None):
        proxy = getattr(instance, self.objectName)
        return proxy[self.key]

    def __set__(self, instance, value):
        proxy = getattr(instance, self.objectName)
        proxy[self.key] = value


S = TypeVar("S", bound=WriterState)


def new_initial_state(klass: Type[S], raw_state: dict) -> S:
    """
    Initializes the initial state of the application and makes it globally accessible.

    The class used for the initial state must be a subclass of WriterState.

    >>> class MyState(WriterState):
    >>>     pass
    >>>
    >>> initial_state = new_initial_state(MyState, {})
    """
    global initial_state
    if raw_state is None:
        raw_state = {}

    initial_state = klass(raw_state)

    return initial_state


"""
This variable contains the list of properties calculated for each class 
that inherits from State.

This mechanic allows Writer Framework to subscribe to mutations that trigger 
these properties when loading an application.
"""
calculated_properties_per_state_type: Dict[Type[State], List[str]] = {}


def writerproperty(path: Union[str, List[str]]):
    """
    Mechanism for declaring a calculated property whenever an attribute changes
    in the state of the Writer Framework application.

    >>> class MyState(wf.WriterState):
    >>>     counter: int
    >>>
    >>>     @wf.property("counter")
    >>>     def double_counter(self):
    >>>         return self.counter * 2

    This mechanism also supports a calculated property that depends on several dependencies.

    >>> class MyState(wf.WriterState):
    >>>     counterA: int
    >>>     counterB: int
    >>>
    >>>     @wf.property(["counterA", "counterB"])
    >>>     def counter_sum(self):
    >>>         return self.counterA + self.counterB
    """

    class Property:
        def __init__(self, func):
            self.func = func
            self.initialized = False
            self.property_name = None

        def __call__(self, *args, **kwargs):
            return self.func(*args, **kwargs)

        def __set_name__(self, owner: Type[State], name: str):
            """
            Saves the calculated properties when loading a State class.
            """
            if owner not in calculated_properties_per_state_type:
                calculated_properties_per_state_type[owner] = []

            calculated_properties_per_state_type[owner].append(name)
            self.property_name = name

        def __get__(self, instance: State, cls):
            """
            This mechanism retrieves the property instance.
            """
            args = inspect.getfullargspec(self.func)
            if len(args.args) > 1:
                logging.warning(
                    f"Wrong signature for calculated property '{instance.__class__.__name__}:{self.property_name}'. It must declare only self argument."
                )
                return None

            if self.initialized is False:
                instance.calculated_property(
                    property_name=self.property_name, path=path, handler=self.func
                )
                self.initialized = True

            return self.func(instance)

    return Property


def session_verifier(func: Callable) -> Callable:
    """
    Decorator for marking session verifiers.
    """

    def wrapped(*args, **kwargs):
        pass

    session_manager.add_verifier(func)
    return wrapped


def get_session() -> Optional[WriterSession]:
    """
    Retrieves the current session.

    This function works exclusively in the context of a request.
    """
    req = _current_request.get()
    if req is None:
        return None

    session_id = req.session_id
    session = session_manager.get_session(session_id)
    if not session:
        return None

    return session


def reset_base_component_tree() -> None:
    """
    Reset the base component tree to zero

    (use mainly in tests)
    """
    global base_component_tree
    base_component_tree = core_ui.build_base_component_tree()


def _clone_mutation_subscriptions(
    session_state: State, app_state: State, root_state: Optional["State"] = None
) -> None:
    """
    clone subscriptions on mutations between the initial state of the application and the state created for the session

    >>> state = wf.init_state({"counter": 0})
    >>> state.subscribe_mutation("counter", lambda state: print(state["counter"]))

    >>> session_state = state.get_clone()

    :param session_state:
    :param app_state:
    :param root_state:
    """
    state_proxy_app = app_state._state_proxy
    state_proxy_session = session_state._state_proxy

    state_proxy_session.local_mutation_subscriptions = []

    _root_state = root_state if root_state is not None else session_state
    for mutation_subscription in state_proxy_app.local_mutation_subscriptions:
        new_mutation_subscription = copy.copy(mutation_subscription)
        new_mutation_subscription.state = (
            _root_state if new_mutation_subscription.type == "subscription" else session_state
        )
        session_state._state_proxy.local_mutation_subscriptions.append(new_mutation_subscription)


def parse_state_variable_expression(p: str):
    r"""
    Parses a state variable expression into a list of parts.

    >>> parse_state_variable_expression("a.b.c")
    >>> ["a", "b", "c"]

    >>> parse_state_variable_expression("a\.b.c")
    >>> ["a.b", "c"]
    """
    parts = []
    it = 0
    last_split = 0
    while it < len(p):
        if p[it] == "\\":
            it += 2
        elif p[it] == ".":
            new_part = p[last_split:it]
            parts.append(new_part.replace("\\.", "."))

            last_split = it + 1
            it += 1
        else:
            it += 1

    new_part = p[last_split : len(p)]
    parts.append(new_part.replace("\\.", "."))
    return parts


async def _async_wrapper_internal(callable_handler: Callable, arg_values: List[Any]) -> Any:
    result = await callable_handler(*arg_values)
    return result


def _event_handler_session_info() -> Dict[str, Any]:
    """
    Returns the session information for the current event handler.

    This information is exposed in the session parameter of a handler

    """
    current_session = get_session()
    session_info: Dict[str, Any] = {}
    if current_session is not None:
        session_info["id"] = current_session.session_id
        session_info["cookies"] = current_session.cookies
        session_info["headers"] = current_session.headers
        session_info["userinfo"] = current_session.userinfo or {}

    return session_info


def _event_handler_ui_manager():
    from writer import PROPER_UI_INIT, _get_ui_runtime_error_message

    if PROPER_UI_INIT:
        from writer.ui import WriterUIManager

        return WriterUIManager()
    else:
        raise RuntimeError(_get_ui_runtime_error_message())


def _deserialize_bigint_format(payload: Optional[Union[dict, list]]):
    """
    Decodes the payload of a big int serialization

    >>> _deserialize_bigint_format({"bigint": "12345678901234567890n"})
    >>> {"bigint" : 12345678901234567890}

    It support escape character on bigint matching format.

    >>> _deserialize_bigint_format({"bigint": '12345678901234567890\n'})
    >>> {"bigint" : "12345678901234567890n"}

    It support nested structure

    >>> _deserialize_bigint_format({
    >>>     "record": {
    >>>         "bigint": '12345678901234567890n'
    >>>     }
    >>> })
    >>> {
    >>>     "record": {
    >>>         "bigint": '12345678901234567890'
    >>>     }
    >>> }
    :param payload:
    :return:
    """
    if isinstance(payload, dict):
        for elt in payload.keys():
            if isinstance(payload[elt], str) and len(payload[elt]) > 0 and payload[elt][-1] == "n":
                if payload[elt][:-1].isdigit():
                    payload[elt] = int(payload[elt][:-1])
                else:
                    unescape_payload = unescape_bigint_matching_string(payload[elt])
                    payload[elt] = unescape_payload

            if isinstance(payload[elt], dict) or isinstance(payload[elt], list):
                _deserialize_bigint_format(payload[elt])
    elif isinstance(payload, list):
        for elt in range(len(payload)):
            if isinstance(payload[elt], str) and payload[elt][-1] == "n":
                if payload[elt][:-1].isdigit():
                    payload[elt] = int(payload[elt][:-1])
                else:
                    unescape_payload = unescape_bigint_matching_string(payload[elt])
                    payload[elt] = unescape_payload
            if isinstance(payload[elt], dict) or isinstance(payload[elt], list):
                _deserialize_bigint_format(payload[elt])

    return payload


def _type_match_dict(expected_type: Type):
    """
    Checks if the expected type expect a dictionary type

    >>> _type_match_dict(dict) # True
    >>> _type_match_dict(int) # False
    >>> _type_match_dict(Dict[str, Any]) # True

    >>> class SpecifcDict(TypedDict):
    >>>     a: str
    >>>     b: str
    >>>
    >>> _type_match_dict(SpecifcDict) # True
    """
    if (
        expected_type is not None
        and inspect.isclass(expected_type)
        and issubclass(expected_type, dict)
    ):
        return True

    if typing.get_origin(expected_type) == dict:
        return True

    return False


def unescape_bigint_matching_string(string: str) -> str:
    """
    Unescapes a string

    >>> unescape_bigint_matching_string(r"13456\n")
    >>> r"13456n"
    """
    if len(string) == 0:
        return string

    if re.match(r"^\d+\\*n$", string) is None:
        return string

    result = ""
    for i in range(len(string)):
        c = string[-i]
        if c == "\\":
            i += 1
            continue
        result += c

    return result


state_serialiser = StateSerialiser()
initial_state = WriterState()
base_component_tree = core_ui.build_base_component_tree()
session_manager: SessionManager = SessionManager()



from writer.abstract import register_abstract_template
from writer.blocks.base_block import WriterBlock
from writer.ss_types import AbstractTemplate, WriterConfigurationError

DEFAULT_MODEL = "palmyra-x5"

class WriterChatReplyWithToolConfig(WriterBlock):
    def __init__(self, component, runner, execution_environment):
        super().__init__(component, runner, execution_environment)

    @classmethod
    def register(cls, type: str):
        super(WriterChatReplyWithToolConfig, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="blueprints_node",
                writer={
                    "name": "Chat reply with tool config",
                    "description": "Initializes conversations, adds messages, and generates replies with tool configuration. NOTE: This is a temporary block with MCP tools mocked - tool calls use hardcoded mock responses instead of making actual HTTP requests.",
                    "category": "Writer",
                    "featureFlags": ["mcp"],
                    "fields": {
                        "conversationStateElement": {
                            "name": "Conversation Object",
                            "desc": "The variable that has your conversation object.",
                            "init": "chat",
                            "type": "Binding",
                        },
                        "systemPrompt": {
                            "name": "System prompt",
                            "type": "Text",
                            "control": "Textarea",
                            "default": "",
                            "desc": "A system prompt to set the context for the conversation. Can be left empty if conversation is already initialized in state.",
                        },
                        "message": {
                            "name": "Message",
                            "type": "Object",
                            "init": '{ "role": "user", "content": "Hello" }',
                            "desc": "A message object. Content can be text string or array of objects for multimodal (text + images) with X5 model.",
                            "validator": {
                                "type": "object",
                                "properties": {
                                    "role": {"type": "string"},
                                    "content": {
                                        "oneOf": [
                                            {"type": "string"},
                                            {
                                                "type": "array",
                                                "items": {
                                                    "oneOf": [
                                                        {
                                                            "type": "object",
                                                            "properties": {
                                                                "type": {
                                                                    "type": "string",
                                                                    "enum": ["text"],
                                                                },
                                                                "text": {"type": "string"},
                                                            },
                                                            "required": ["type", "text"],
                                                            "additionalProperties": False,
                                                        },
                                                        {
                                                            "type": "object",
                                                            "properties": {
                                                                "type": {
                                                                    "type": "string",
                                                                    "enum": ["image_url"],
                                                                },
                                                                "image_url": {
                                                                    "type": "object",
                                                                    "properties": {
                                                                        "url": {"type": "string"}
                                                                    },
                                                                    "required": ["url"],
                                                                    "additionalProperties": False,
                                                                },
                                                            },
                                                            "required": ["type", "image_url"],
                                                            "additionalProperties": False,
                                                        },
                                                    ]
                                                },
                                            },
                                        ]
                                    },
                                },
                                "additionalProperties": False,
                            },
                        },
                        "generateReply": {
                            "name": "Generate reply",
                            "type": "Boolean",
                            "default": "yes",
                            "desc": "If set to 'yes', the block will generate a reply from the model after adding the message. If set to 'no', it will only add the message to the conversation.",
                            "validator": {
                                "type": "boolean",
                            },
                        },
                        "initModelId": {
                            "name": "Initial model",
                            "type": "Model Id",
                            "default": DEFAULT_MODEL,
                        },
                        "initTemperature": {
                            "name": "Initial temperature",
                            "type": "Number",
                            "default": "0.7",
                            "validator": {
                                "type": "number",
                                "minimum": 0,
                                "maximum": 2,
                            },
                        },
                        "initMaxTokens": {
                            "name": "Initial max tokens",
                            "type": "Number",
                            "default": "1024",
                            "validator": {
                                "type": "number",
                                "minimum": 1,
                                "maximum": 16384,
                            },
                        },
                        "useStreaming": {
                            "name": "Use streaming",
                            "type": "Boolean",
                            "default": "yes",
                            "desc": "If set to 'yes', the block will stream the reply as it is generated. If set to 'no', it will wait for the entire reply to be generated before returning.",
                            "validator": {
                                "type": "boolean",
                            },
                        },
                        "tools": {
                            "name": "Tools",
                            "type": "Tools",
                            "default": "{}",
                            "init": "",
                            "category": "Tools",
                        },
                        "toolConfig": {
                            "name": "Tool config",
                            "type": "Object",
                            "default": "{}",
                            "desc": "Additional configuration for tools.",
                            "validator": {
                                "type": "object",
                                "additionalProperties": True,
                            },
                        },
                    },
                    "outs": {
                        "tools": {
                            "name": "Tools",
                            "field": "tools",
                            "description": "Run associated tools.",
                            "style": "dynamic",
                        },
                        "success": {
                            "name": "Success",
                            "description": "The reply was generated successfully.",
                            "style": "success",
                        },
                        "error": {
                            "name": "Error",
                            "description": "There was an error generating the reply.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def _make_callable(self, tool_name: str, app_id: str, function_name: str):
        def callable(**args):
            import json
            import os
            import uuid

            import httpx

            from writer.core import get_session
            current_session = get_session()
            
            org_id = None
            user_id = None
            
            if current_session and current_session.headers:
                org_id = current_session.headers.get("x-organization-id")
                user_id = current_session.headers.get("x-user-id")
            
            if not org_id:
                org_id = os.getenv("WRITER_ORG_ID")
            
            if not org_id:
                raise ValueError("Organization ID not found. Set x-organization-id header or WRITER_ORG_ID environment variable.")
            
            if not user_id:
                user_id = os.getenv("WRITER_USER_ID", "1")
            
            try:
                user_id = int(user_id)
            except (ValueError, TypeError):
                user_id = 1
            
            # url = f"http://localhost:8001/api/v1/namespaces/default/services/mcp-gateway:80/proxy/api/private/mcp-gateway/organization/1/mcp"
            # auth_token_header = "X-Auth-Token-Data-User-Individual"
            # auth_token_data = json.dumps({
            #     "userId": user_id,
            #     "authType": "Password",
            #     "confirmed": True
            # }).encode()
            #
            # headers = {
            #     "Content-Type": "application/json",
            #     auth_token_header: auth_token_data,
            # }
            #
            # jsonrpc_request = {
            #     "jsonrpc": "2.0",
            #     "id": str(uuid.uuid4()),
            #     "method": "tools/call",
            #     "params": {
            #         "name": function_name,
            #         "arguments": args
            #     }
            # }
            
            result = {
                'id': '2c22bbee-65eb-4471-bd05-4e1be617db74',
                'jsonrpc': '2.0',
                'result': {
                    'content': [{
                        'text': '[\n  45873434,\n  45873113,\n  45785840\n]',
                        'type': 'text'
                    }]
                }
            }
            
            # if "error" in result:
            #     error = result["error"]
            #     error_message = error.get("message", "Unknown MCP error") if isinstance(error, dict) else str(error)
            #     raise RuntimeError(f"MCP gateway API returned error: {error_message}")
            
            result_data = result.get("result", {})
            
            content_items = result_data.get("content", [])
            if content_items:
                return_value = []
                for item in content_items:
                    if isinstance(item, dict):
                        if item.get("type") == "text":
                            return_value.append(item.get("text", ""))
                        elif item.get("type") == "image":
                            return_value.append(f"[Image: {item.get('mimeType', 'image/png')}]")
                        else:
                            return_value.append(str(item))
                    else:
                        return_value.append(str(item))
                return_value = "\n".join(return_value) if return_value else ""
            else:
                return_value = result_data.get("result") or result_data.get("content") or result_data
            
            return return_value

        return callable

    def run(self):
        try:
            import writer.ai

            conversation_state_element = self._get_field("conversationStateElement", required=True)
            message = self._get_field("message", as_json=True)
            generate_reply = self._get_field("generateReply", False, "yes") == "yes"

            system_prompt = self._get_field("systemPrompt", False, default_field_value=None)
            init_model_id = self._get_field("initModelId", False, default_field_value=DEFAULT_MODEL)
            try:
                init_temperature = float(self._get_field("initTemperature", False, "0.7"))
                init_max_tokens = int(self._get_field("initMaxTokens", False, "1024"))
            except ValueError as e:
                raise WriterConfigurationError(f"Invalid numeric value in configuration: {e}")
            use_streaming = self._get_field("useStreaming", False, "yes") == "yes"
            tools_raw = self._get_field("tools", True)

            tools = []

            conversation = self.evaluator.evaluate_expression(
                conversation_state_element, self.instance_path, self.execution_environment
            )

            if conversation is None:
                config = {
                    "temperature": init_temperature,
                    "model": init_model_id,
                    "max_tokens": init_max_tokens,
                }
                conversation = writer.ai.Conversation(
                    prompt_or_history=system_prompt, config=config
                )
                self._set_state(conversation_state_element, conversation)
            elif not isinstance(conversation, writer.ai.Conversation):
                raise WriterConfigurationError(
                    "The state element specified doesn't contain a Conversation."
                )

            if message not in (None, {}, ""):
                writer.ai.Conversation.validate_message(message)
                conversation += message
                self._set_state(conversation_state_element, conversation)

            if not generate_reply:
                self.result = ""
                self.outcome = "success"
                return

            for tool_name, tool_raw in tools_raw.items():
                if not isinstance(tool_raw, dict):
                    continue
                
                if not isinstance(tool_name, str) or not tool_name:
                    continue
                
                tool_type = tool_raw.get("type")
                if tool_type != "mcp":
                    continue
                
                try:
                    app_id = tool_raw.get("appId")
                    function_name = tool_raw.get("functionName")
                    
                    if not app_id:
                        continue
                    
                    if not function_name:
                        continue
                    
                    mcp_function = tool_raw.get("function", {})
                    description = mcp_function.get("description", "")
                    parameters = mcp_function.get("parameters", {})
                    
                    if parameters:
                        if not isinstance(parameters, dict):
                            continue
                        
                        if "properties" in parameters:
                            parameters = parameters.get("properties", {})
                        elif "type" not in parameters:
                            parameters = {
                                "type": "object",
                                "properties": parameters
                            }

                    tool = writer.ai.FunctionTool(
                        type="function",
                        name=tool_name,
                        description=description,
                        callable=self._make_callable(tool_name, app_id, function_name),
                        parameters=parameters or {},
                    )
                    tools.append(tool)

                except Exception:
                    continue
            
            if tools_raw and len(tools) == 0:
                raise WriterConfigurationError(
                    "No valid tools could be created from the provided tools configuration. "
                    "Please check that tools have the required fields and valid structure."
                )

            msg = ""
            if not use_streaming:
                reply = conversation.complete(tools=tools)
                msg = reply.get("content") or ""
                conversation += reply
                self._set_state(conversation_state_element, conversation)
            else:
                for chunk in conversation.stream_complete(tools=tools):
                    if chunk.get("content") is None:
                        chunk["content"] = ""
                    msg += chunk.get("content")
                    conversation += chunk
                    self._set_state(conversation_state_element, conversation)
            self.result = msg
            self._set_state(conversation_state_element, conversation)
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e




import argparse
import asyncio
import os
import shutil
import stat
import tempfile
import threading
import time
from collections import defaultdict
from datetime import datetime, timedelta
from pathlib import Path
from threading import Timer
from typing import Any, Dict, List, Optional, Set, Union

from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer
from watchdog.observers.api import BaseObserver


class OperationQueues:
    def __init__(self) -> None:
        self.files_to_update: Dict[str, str] = {}  # Maps dest paths to their source paths
        self.files_to_delete: Set[str] = set()
        self.dirs_to_create: Set[str] = set()
        self.dirs_to_delete: Set[str] = set()
        self.lock = threading.Lock()

    def has_pending_operations(self) -> bool:
        with self.lock:
            return (
                len(self.files_to_update) > 0 or
                len(self.files_to_delete) > 0 or
                len(self.dirs_to_create) > 0 or
                len(self.dirs_to_delete) > 0
            )

    def clear_all(self) -> None:
        with self.lock:
            self.files_to_update.clear()
            self.files_to_delete.clear()
            self.dirs_to_create.clear()
            self.dirs_to_delete.clear()


class FolderSyncHandler(FileSystemEventHandler):
    def __init__(self, src_dir: str, dest_dir: str, verbose: bool, operation_queues: OperationQueues) -> None:
        self.src_dir = Path(src_dir).resolve()
        self.dest_dir = Path(dest_dir).resolve()
        self.verbose = verbose
        self.operation_queues = operation_queues

    def log(self, *args) -> None:
        if self.verbose:
            print(*args)

    def rel(self, path: str) -> str:
        return str(Path(path).relative_to(self.src_dir))

    def dest_path(self, src_path: str) -> str:
        rel_path = self.rel(src_path)
        return str(self.dest_dir / rel_path)

    def on_created(self, event) -> None:
        if event.is_directory:
            with self.operation_queues.lock:
                self.operation_queues.dirs_to_create.add(event.src_path)
            self.log("Queued dir add:", self.rel(event.src_path))
        else:
            dest = self.dest_path(event.src_path)
            with self.operation_queues.lock:
                self.operation_queues.files_to_update[dest] = event.src_path
            self.log("Queued add:", self.rel(event.src_path))

    def on_modified(self, event) -> None:
        if not event.is_directory:
            dest = self.dest_path(event.src_path)
            with self.operation_queues.lock:
                self.operation_queues.files_to_update[dest] = event.src_path
            self.log("Queued change:", self.rel(event.src_path))

    def on_deleted(self, event) -> None:
        if event.is_directory:
            with self.operation_queues.lock:
                self.operation_queues.dirs_to_delete.add(event.src_path)
            self.log("Queued dir remove:", self.rel(event.src_path))
        else:
            dest = self.dest_path(event.src_path)
            with self.operation_queues.lock:
                self.operation_queues.files_to_delete.add(event.src_path)
                # Remove from update queue if present
                self.operation_queues.files_to_update.pop(dest, None)
            self.log("Queued remove:", self.rel(event.src_path))

    def on_moved(self, event) -> None:
        # Handle moves as delete + create
        self.on_deleted(type('MockEvent', (), {'src_path': event.src_path, 'is_directory': event.is_directory})())
        self.on_created(type('MockEvent', (), {'src_path': event.dest_path, 'is_directory': event.is_directory})())


class FileBuffering:
    def __init__(self, dest_dir: str, *, src_dir: Optional[str] = None, verbose: bool = False, interval: int = 10) -> None:
        self.path = src_dir if src_dir is not None else tempfile.mkdtemp()
        self.src_dir = Path(self.path).resolve()
        self.dest_dir = Path(dest_dir).resolve()
        self.verbose = verbose
        self.sync_interval = interval
        self.operation_queues = OperationQueues()
        self.is_sync_in_progress = False
        self.observer: Optional[BaseObserver] = None
        self.sync_timer: Optional[Timer] = None
        self.running = False
        self.last_comparison_time: datetime = datetime.now()
        self.check_interval = 10
        self.check_timer: Optional[Timer] = None

    def log(self, *args) -> None:
        if self.verbose:
            print(*args)

    def rel(self, path: str) -> str:
        return str(Path(path).relative_to(self.src_dir))

    def dest_path(self, src_path: str) -> str:
        rel_path = self.rel(src_path)
        return str(self.dest_dir / rel_path)

    def should_copy(self, src: str, dest: str) -> bool:
        try:
            src_path = Path(src)
            dest_path = Path(dest)
            
            if not dest_path.exists():
                return True
                
            src_stat = src_path.stat()
            dest_stat = dest_path.stat()
            
            # Compare size
            if src_stat.st_size != dest_stat.st_size:
                return True
            
            # Compare modification time (with 1 second tolerance)
            if src_stat.st_mtime > dest_stat.st_mtime + 1:
                return True
            
            # Compare content
            try:
                with open(src, 'rb') as f1, open(dest, 'rb') as f2:
                    return f1.read() != f2.read()
            except Exception:
                return True
                
        except Exception:
            return True

    def ensure_dir(self, path: str) -> None:
        Path(path).mkdir(parents=True, exist_ok=True)

    def copy_file(self, src: str, dest: str) -> None:
        dest_path = Path(dest)
        dest_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(src, dest)

    def remove_path(self, path: str) -> None:
        path_obj = Path(path)
        if path_obj.exists():
            if path_obj.is_dir():
                shutil.rmtree(path_obj)
            else:
                path_obj.unlink()

    def init(self) -> None:
        """Initialize by copying all files from destination to source directory."""
        self.log(f"Initializing: copying files from {self.dest_dir} to {self.src_dir}")
        
        # Check if destination directory exists
        if not self.dest_dir.exists():
            self.log(f"Destination directory {self.dest_dir} doesn't exist yet. Nothing to copy.")
            return
        
        def copy_recursively(relative_path: str = ""):
            full_dest = self.dest_dir / relative_path
            full_src = self.src_dir / relative_path
            
            # Ensure the directory exists in source
            self.ensure_dir(str(full_src))
            
            try:
                for entry in full_dest.iterdir():
                    rel_path = str(Path(relative_path) / entry.name) if relative_path else entry.name
                    dest_path = self.dest_dir / rel_path
                    src_path = self.src_dir / rel_path
                    
                    if entry.is_dir():
                        copy_recursively(rel_path)
                    elif entry.is_file():
                        self.copy_file(str(dest_path), str(src_path))
                        self.log(f"Copied to source: {rel_path}")
            except PermissionError as e:
                print(f"Permission error accessing {full_dest}: {e}")
            except Exception as e:
                print(f"Error processing {full_dest}: {e}")
        
        # Execute the recursive copy
        copy_recursively()
        self.log("Initialization complete. All files copied to temporary directory.")
        
    def initial_sync(self) -> None:
        self.log("Starting initial sync...")
        
        def sync_recursively(relative: str = ""):
            full_src = self.src_dir / relative
            full_dest = self.dest_dir / relative
            
            self.ensure_dir(str(full_dest))
            
            try:
                for entry in full_src.iterdir():
                    rel_path = str(Path(relative) / entry.name) if relative else entry.name
                    src_path = self.src_dir / rel_path
                    dst_path = self.dest_dir / rel_path
                    
                    if entry.is_dir():
                        sync_recursively(rel_path)
                    elif entry.is_file():
                        if self.should_copy(str(src_path), str(dst_path)):
                            self.copy_file(str(src_path), str(dst_path))
                            self.log("Copied:", rel_path)
                        else:
                            self.log("Skipped (up to date):", rel_path)
            except PermissionError as e:
                print(f"Permission error accessing {full_src}: {e}")
            except Exception as e:
                print(f"Error processing {full_src}: {e}")
        
        # Run synchronously since it's the initial sync
        sync_recursively()

    def process_batched_operations(self) -> None:
        if self.is_sync_in_progress:
            return
        if not self.operation_queues.has_pending_operations():
            return
        
        self.is_sync_in_progress = True
        self.log(f"Processing batched operations at {datetime.now().isoformat()}")
        
        try:
            with self.operation_queues.lock:
                # Process directory creations first
                if self.operation_queues.dirs_to_create:
                    self.log(f"Creating {len(self.operation_queues.dirs_to_create)} directories")
                    for dir_path in self.operation_queues.dirs_to_create.copy():
                        try:
                            self.log("Dir add:", self.rel(dir_path))
                            to = self.dest_path(dir_path)
                            self.ensure_dir(to)
                        except Exception as e:
                            print(f"Error creating directory {dir_path}: {e}")
                    self.operation_queues.dirs_to_create.clear()
                
                # Process file updates
                if self.operation_queues.files_to_update:
                    self.log(f"Updating {len(self.operation_queues.files_to_update)} files")
                    for dest_path, src_path in self.operation_queues.files_to_update.copy().items():
                        try:
                            if self.should_copy(src_path, dest_path):
                                self.log("File update:", self.rel(src_path))
                                self.copy_file(src_path, dest_path)
                        except Exception as e:
                            print(f"Error updating file {src_path} to {dest_path}: {e}")
                    self.operation_queues.files_to_update.clear()
                
                # Process file deletions
                if self.operation_queues.files_to_delete:
                    self.log(f"Deleting {len(self.operation_queues.files_to_delete)} files")
                    for file_path in self.operation_queues.files_to_delete.copy():
                        try:
                            self.log("File remove:", self.rel(file_path))
                            to = self.dest_path(file_path)
                            self.remove_path(to)
                        except Exception as e:
                            print(f"Error deleting file {file_path}: {e}")
                    self.operation_queues.files_to_delete.clear()
                
                # Process directory deletions last
                if self.operation_queues.dirs_to_delete:
                    self.log(f"Deleting {len(self.operation_queues.dirs_to_delete)} directories")
                    for dir_path in self.operation_queues.dirs_to_delete.copy():
                        try:
                            self.log("Dir remove:", self.rel(dir_path))
                            to = self.dest_path(dir_path)
                            self.remove_path(to)
                        except Exception as e:
                            print(f"Error deleting directory {dir_path}: {e}")
                    self.operation_queues.dirs_to_delete.clear()
                    
        except Exception as error:
            print(f"Error processing batch: {error}")
        finally:
            self.is_sync_in_progress = False
        
        # Compare directories after processing
        self.compare_directories(str(self.src_dir), str(self.dest_dir), self.verbose)

    def schedule_batch_processing(self) -> None:
        if not self.running:
            return
            
        self.process_batched_operations()

        
        # Schedule next processing
        self.sync_timer = threading.Timer(self.sync_interval, self.schedule_batch_processing)
        if self.sync_timer:  # This check is redundant but satisfies mypy
            self.sync_timer.start()

    def schedule_sanity_check(self) -> None:

        if self.last_comparison_time < datetime.now() - timedelta(seconds=5):
            self.compare_directories(str(self.src_dir), str(self.dest_dir), self.verbose)

        self.check_timer = threading.Timer(self.check_interval, self.schedule_sanity_check)
        if self.check_timer:  # This check is redundant but satisfies mypy
            self.check_timer.start()



    def get_all_files(self, directory: str, base_dir: str) -> Dict[str, str]:
        files = {}
        dir_path = Path(directory)
        base_path = Path(base_dir)
        
        def traverse_dir(current_path: Path):
            try:
                for entry in current_path.iterdir():
                    if entry.is_dir():
                        traverse_dir(entry)
                    elif entry.is_file():
                        relative_path = str(entry.relative_to(base_path))
                        files[relative_path] = str(entry)
            except PermissionError:
                pass
        
        traverse_dir(dir_path)
        return files

    def files_have_same_content(self, file1: str, file2: str) -> bool:
        try:
            with open(file1, 'rb') as f1, open(file2, 'rb') as f2:
                return f1.read() == f2.read()
        except Exception as e:
            if self.verbose:
                print(f"Error comparing files {file1} and {file2}: {e}")
            return False

    def compare_directories(self, dir1: str, dir2: str, verbose: bool = False) -> List[Dict[str, str]]:
        divergent_files: List[Dict[str, str]] = []
        self.last_comparison_time = datetime.now()
        
        def log_verbose(*args):
            if verbose:
                print(*args)
        
        # Get all files from both directories
        dir1_files = self.get_all_files(dir1, dir1)
        dir2_files = self.get_all_files(dir2, dir2)
        
        # Compare files that exist in both directories
        for relative_path, full_path1 in dir1_files.items():
            full_path2 = dir2_files.get(relative_path)
            
            if not full_path2:
                # File exists in dir1 but not in dir2
                divergent_files.append({'path': relative_path, 'reason': 'missing in target'})
                dest = self.dest_path(full_path1)
                with self.operation_queues.lock:
                    self.operation_queues.files_to_update[dest] = full_path1
                log_verbose(f"File only in source: {relative_path}")
            else:
                # File exists in both, compare content
                if not self.files_have_same_content(full_path1, full_path2):
                    divergent_files.append({'path': relative_path, 'reason': 'content differs'})
                    dest = self.dest_path(full_path1)
                    with self.operation_queues.lock:
                        self.operation_queues.files_to_update[dest] = full_path1
                    log_verbose(f"Content differs: {relative_path}")
        
        # Check for files in dir2 that don't exist in dir1
        for relative_path in dir2_files.keys():
            if relative_path not in dir1_files:
                divergent_files.append({'path': relative_path, 'reason': 'missing in source'})
                with self.operation_queues.lock:
                    self.operation_queues.files_to_delete.add(str(self.src_dir / relative_path))
                log_verbose(f"File only in target: {relative_path}")
        
        return divergent_files

    def watch_changes(self) -> None:
        """
        Set up file system monitoring and start the periodic sync process.
        This method returns immediately after setting up the background watchers.
        """
        handler = FolderSyncHandler(str(self.src_dir), str(self.dest_dir), self.verbose, self.operation_queues)
        self.observer = Observer()
        if self.observer:  # This check is redundant but satisfies mypy
            self.observer.schedule(handler, str(self.src_dir), recursive=True)
            self.observer.start()
        
        print(f"Watching for changes in {self.src_dir}...")
        print(f"Files will be synchronized every {self.sync_interval} seconds")
        
        # Start the batch processing scheduler
        self.running = True
        self.schedule_batch_processing()
        self.schedule_sanity_check()

    def stop(self) -> None:
        self.running = False
        if self.sync_timer:
            self.sync_timer.cancel()
            self.sync_timer.join()
            self.sync_timer = None
        if self.check_timer:
            self.check_timer.cancel()
            self.check_timer.join()
            self.check_timer = None
        if self.observer:
            self.observer.stop()
            self.observer.join()




from typing import Any

from writer.abstract import register_abstract_template
from writer.blocks.base_block import WriterBlock
from writer.ss_types import AbstractTemplate, WriterConfigurationError


class WriterAddToKG(WriterBlock):

    @classmethod
    def register(cls, type: str):
        super(WriterAddToKG, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="blueprints_node",
            writer={
                "name": "Add to Knowledge Graph",
                "description": "Adds structured information to the knowledge graph. Use for storing facts AI can reference.",
                "category": "Writer",
                "fields": {
                    "graphId": {
                        "name": "Graph Id",
                        "type": "Graph Id",
                        "desc": "The id for an existing knowledge graph. It has a UUID format.",
                        "validator": {
                            "type": "string",
                            "format": "uuid",
                        }
                    },
                    "files": {
                        "name": "Files",
                        "type": "Object",
                        "default": "[]",
                        "desc": "A list of files to be uploaded and added to the knowledge graph. You can use files uploaded via the File input component or specify dictionaries with data, type and name.",
                        "validator": {
                            "type": "array",
                        }
                    },
                    "urls": {
                        "name": "URLs",
                        "type": "Object",
                        "default": "[]",
                        "desc": "A list of URLs to be added to the knowledge graph. Web content from these URLs will be indexed.",
                        "validator": {
                            "type": "array",
                            "items": {
                                "type": "string",
                                "format": "uri",
                            }
                        }
                    },
                },
                "outs": {
                    "success": {
                        "name": "Success",
                        "description": "Content was added to the knowledge graph.",
                        "style": "success",
                    },
                    "error": {
                        "name": "Error",
                        "description": "There was an error adding content to the knowledge graph.",
                        "style": "error",
                    },
                },
            }
        ))

    def _get_prepared_file(self, raw_file: Any):
        if not isinstance(raw_file, dict):
            raise WriterConfigurationError("Files must be dictionaries and contain `data`, `type` and `name` attributes.")

        if "data" not in raw_file or "type" not in raw_file or "name" not in raw_file:
            raise WriterConfigurationError("A file specified as a dictionary must contain `data`, `type` and `name` attributes.")

        return raw_file

    def run(self):
        try:
            import writer.ai

            graph_id = self._get_field("graphId", required=True)
            raw_files = self._get_field("files", as_json=True, default_field_value="[]")
            urls = self._get_field("urls", as_json=True, default_field_value="[]")
            prepared_files = []

            if not isinstance(raw_files, list):
                raise WriterConfigurationError("Files must be a list.")
            
            if not isinstance(urls, list):
                raise WriterConfigurationError("URLs must be a list.")
            
            if not raw_files and not urls:
                raise WriterConfigurationError("At least one file or URL must be provided.")

            for raw_file in raw_files:
                prepared_files.append(self._get_prepared_file(raw_file))
                
            graph = writer.ai.retrieve_graph(graph_id)
            
            # Add files to the graph
            for prepared_file in prepared_files:
                file = writer.ai.upload_file(prepared_file.get("data"),
                                             prepared_file.get("type"),
                                             prepared_file.get("name"))
                graph.add_file(file)
            
            # Add URLs to the graph using SDK 2.3.0 feature
            for url in urls:
                if not isinstance(url, str):
                    raise WriterConfigurationError(f"URL must be a string, got {type(url)}")
                graph.add_url(url)

            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



import re

from writer.abstract import register_abstract_template
from writer.blocks.base_block import WriterBlock
from writer.ss_types import AbstractTemplate, WriterConfigurationError

ALLOWED_CHARS = re.compile(r'^[A-Za-z0-9\-_]+$')

class WriterKeyValueStorage(WriterBlock):
    @classmethod
    def register(cls, type: str):
        super(WriterKeyValueStorage, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="blueprints_node",
                writer={
                    "name": "Key-Value Storage",
                    "description": "Stores data between sessions. Uses unique keys (names) to identify the data. Keys can only contain alphanumeric characters, underscores, and hyphens.",
                    "category": "Writer",
                    "fields": {
                        "action": {
                            "name": "Action",
                            "type": "Text",
                            "description": "What action to perform on the data (save, get, delete, list all keys).",
                            "options": {
                                "Save": "Save",
                                "Get": "Get",
                                "Delete": "Delete",
                                "List keys": "List keys",
                            },
                            "default": "Save",
                        },
                        "key": {
                            "name": "Key",
                            "type": "Text",
                            "description": "Unique identifier of your data that will be used to retrieve, update and delete it.",
                        },
                        "valueType": {
                            "name": "Value type",
                            "type": "Text",
                            "description": "Specify whether to interpret the value as plain text or JSON.",
                            "options": {
                                "text": "Plain text",
                                "JSON": "JSON",
                            },
                            "default": "text",
                        },
                        "value": {
                            "name": "Value",
                            "type": "Text",
                            "description": "Data that you want to store.",
                            "control": "Textarea",
                        },
                    },
                    "outs": {
                        "success": {
                            "name": "Success",
                            "description": "The request was successful.",
                            "style": "success",
                        },
                        "error": {
                            "name": "Error",
                            "description": "The request wasn't successful.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def run(self):
        try:
            self.result = self._execute_action()
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e

    def _execute_action(self):
        from writer.journal import JOURNAL_KEY_PREFIX
        from writer.keyvalue_storage import KeyValueStorage

        action = self._get_field("action", default_field_value="Save")

        with self.acquire_httpx_client() as client:
            writer_kv_storage = KeyValueStorage(client=client)

            if action == "List keys":
                response = writer_kv_storage.get_data_keys()
                return [key for key in response if not key.startswith(JOURNAL_KEY_PREFIX)]

            key = self._get_field("key", required=True)
            if not ALLOWED_CHARS.fullmatch(key):
                raise WriterConfigurationError("Key can only contain alphanumeric characters, underscores and hyphens")

            if action == "Save":
                value_type = self._get_field("valueType")
                value = self._get_field("value", as_json=value_type == "JSON")
                return writer_kv_storage.save(key, value)
            if action == "Get":
                return writer_kv_storage.get(key, type_="data")["data"]
            if action == "Delete":
                return writer_kv_storage.delete(key)

            raise WriterConfigurationError(f"Unknown action for the Key-Value Storage: {action}")



from writer.abstract import register_abstract_template
from writer.blocks.base_trigger import BlueprintTrigger
from writer.ss_types import AbstractTemplate


class CronTrigger(BlueprintTrigger):
    @classmethod
    def register(cls, type: str):
        super(CronTrigger, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="blueprints_node",
                writer={
                    "name": "Cron Trigger",
                    "description": "Triggers an event on a schedule based on a cron expression.",
                    "category": "Triggers",
                    "featureFlags": ["cron_trigger"],
                    "fields": {
                        "cronExpression": {
                            "name": "Cron Expression",
                            "type": "Text",
                            "desc": "The cron expression that defines the schedule (e.g., '0 * * * *' for every hour).",
                        },
                        "timezone": {
                            "name": "Timezone",
                            "type": "Text",
                            "desc": "The timezone for the cron schedule (e.g., 'UTC', 'America/New_York'). Defaults to UTC.",
                            "default": "UTC",
                        },
                        "defaultResult": {
                            "name": "Default result",
                            "type": "Code",
                            "desc": 'The result that is used when the blueprint is triggered from the "Run blueprint" button',
                        },
                    },
                    "outs": {
                        "trigger": {
                            "name": "Trigger",
                            "style": "success",
                        },
                    },
                },
            ),
        )

    def run(self):
        super().run()
        self.outcome = "trigger"



from writer.abstract import register_abstract_template
from writer.blocks.base_block import BlueprintBlock
from writer.ss_types import AbstractTemplate, WriterConfigurationError


class AddToStateList(BlueprintBlock):

    @classmethod
    def register(cls, type: str):
        super(AddToStateList, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="blueprints_node",
            writer={
                "name": "Add to state list",
                "description": "Adds a new item to a list in the Agent's state. Useful for tracking multiple values over time.",
                "category": "Other",
                "fields": {
                    "element": {
                        "name": "Link Variable",
                        "type": "Binding",
                        "desc": "Set the variable here and use it across your agent.",
                    },
                    "value": {
                        "name": "Value",
                        "type": "Text",
                        "control": "Textarea"
                    },
                },
                "outs": {
                    "success": {
                        "name": "Success",
                        "description": "The item was added to the list.",
                        "style": "success",
                    },
                    "error": {
                        "name": "Error",
                        "description": "There was an error adding the item to the list.",
                        "style": "error",
                    },
                },
            }
        ))

    def run(self):
        try:
            element_expr = self._get_field("element")
            value = self._get_field("value")

            element = self.evaluator.evaluate_expression(element_expr, self.instance_path, self.execution_environment)

            if not element:
                element = []
            elif not isinstance(element, list):
                raise WriterConfigurationError(f'The state element must be a list. A value of type "{type(element)}" was found.')

            element.append(value)
            self._set_state(element_expr, element)
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



import contextlib
import copy
import logging
import uuid
from contextvars import ContextVar
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union, cast

from pydantic import BaseModel, Field, field_validator, validator
from typing_extensions import TypedDict

from writer.ss_types import ComponentDefinition, ServeMode

current_parent_container: ContextVar[Union["Component", None]] = \
    ContextVar("current_parent_container")

# This variable is thread safe and context safe
_current_component_tree: ContextVar[Union["ComponentTree", None]] = \
    ContextVar("current_component_tree", default=None)


def generate_component_id():
    return str(uuid.uuid4())


class Branch(Enum):
    """
    Enum for the component tree branches that can be created in Writer Framework

    * bmc: builder managed component
    * initial_cmc: code managed component
    * session: session managed component

    This enum should be used only in the module core_ui.py.
    """
    bmc = "bmc"
    initial_cmc = "initial_cmc"
    session_cmc = "session_cmc"


class VisibileFields(TypedDict):
    expression: Union[bool, Literal['custom']]
    binding: str
    reversed: bool

class Component(BaseModel):
    id: str = Field(default_factory=generate_component_id)
    type: str
    content: Dict[str, Any] = Field(default_factory=dict)
    isCodeManaged: Optional[bool] = False
    position: int = 0
    parentId: Optional[str] = None
    handlers: Optional[Dict[str, str]] = None
    visible: Optional[VisibileFields] = None
    binding: Optional[Dict] = None
    outs: Optional[Any] = None
    x: Optional[int] = None
    y: Optional[int] = None

    @field_validator("x", "y", mode="before")
    def cast_float_to_int(cls, v):
        if isinstance(v, float):
            return int(v)
        return v

    def to_dict(self) -> Dict:
        """
        Wrapper for model_dump to ensure backward compatibility.
        """
        return self.model_dump(exclude_none=True)

    def __enter__(self) -> "Component":
        self._token = current_parent_container.set(self)
        return self

    def __exit__(self, *_):
        current_parent_container.reset(self._token)


class ComponentTreeBranch:
    """
    >>> bmc_tree = ComponentTreeBranch(Branch.bmc, False)
    """

    def __init__(self, id: Branch, freeze: bool = False) -> None:
        """

        :param id: the id of the component tree branch (bmc, cmc, session)
        :param freeze: the component list can not be modified
        """
        self.components: Dict[str, Component] = {}
        # Page counter is required to set
        # predefined IDs & keys for code-managed pages
        self.page_counter = 0
        self.component_tree_id = id
        self.freeze = freeze

    def clone(self) -> 'ComponentTreeBranch':
        """
        Clone a component branch into a new one.

        Writer Framework uses this action when it instantiates the component tree attached to the session.
        This ensures complete insulation of the original shaft.

        >>> cloned = bmc_tree.clone()
        :return:
        """
        cloned = ComponentTreeBranch(self.component_tree_id, self.freeze)
        cloned.components = copy.copy(self.components)
        cloned.page_counter = self.page_counter
        return cloned


    def get_component(self, component_id: str) -> Optional[Component]:
        return self.components.get(component_id)

    def attach(self, component: Component) -> None:
        """
        Attaches a component to the main components dictionary of the instance.

        :param component: The component to be attached.
        """
        if self.freeze:
            raise UIError(f"Component tree {self.component_tree_id} is frozen and cannot be modified")

        if (component.id in self.components):
            raise RuntimeWarning(f"Component with ID {component.id} already exists")

        if component.type == "page" and component.id not in self.components:
            self.page_counter += 1

        self.components[component.id] = component

    def ingest(self, serialised_components: Dict[str, Any]) -> None:
        if self.freeze:
            raise UIError(f"Component tree {self.component_tree_id} is frozen and cannot be modified")

        removed_ids = [key for key in self.components if key not in serialised_components]

        for component_id in removed_ids:
            if component_id == "root":
                continue
            self.components.pop(component_id)

        for component_id, sc in serialised_components.items():
            component = Component(**sc)
            self.components[component_id] = component

    def to_dict(self) -> Dict:
        active_components = {}
        for id, component in self.components.items():
            active_components[id] = component.to_dict()

        return active_components



class ComponentTree():

    def __init__(self, tree_branches: List[ComponentTreeBranch]):
        assert len(tree_branches) > 0, "Component tree must have at least one tree branch"
        self.tree_branches = tree_branches
        self.updated = False

    @property
    def components(self) -> Dict[str, Component]:
        trees = reversed(self.tree_branches)
        all_components = {}
        for tree in trees:
            all_components.update(tree.components)
        return all_components

    @property
    def page_counter(self) -> int:
        return sum([tree.page_counter for tree in self.tree_branches])

    def get_component(self, component_id: str) -> Optional[Component]:
        for tree in self.tree_branches:
            component = tree.get_component(component_id)
            if component:
                return component

        return None

    def attach(self, component: Component, tree: Optional[Branch] = None) -> None:
        """
        Attach a component to the first tree branch in the component tree.

        >>> component = Component(id="root", type="root", content={})
        >>> component_tree.attach(component)

        When the tree parameter is given, the component is attached to the corresponding branch.

        >>> component = Component(id="root", type="root", content={})
        >>> component_tree.attach(component, tree=Branch.initial_cmc)

        If the component is associated with a frozen branch, an exception is thrown
        """
        self.updated = True

        _branch = self._tree_branch(tree)
        if _branch is None:
            raise ValueError(f"Invalid tree branch : {tree}")

        cast(ComponentTreeBranch, _branch).attach(component)

    def ingest(self, serialised_components: Dict[str, Any], tree: Optional[Branch] = None) -> None:
        self.updated = True
        _branch = self._tree_branch(tree)
        if _branch is None:
            raise ValueError(f"Invalid tree branch : {tree}")

        cast(ComponentTreeBranch, _branch).ingest(serialised_components)

    def delete_component(self, component_id: str) -> None:
        for tree_branch in self.tree_branches:
            if component_id in tree_branch.components and not tree_branch.freeze:
                self.updated = True
                tree_branch.components.pop(component_id, None)
                return
            elif component_id in tree_branch.components and tree_branch.freeze:
                raise UIError(
                    f"Component with ID '{component_id}' " +
                    "is builder-managed and cannot be removed by app"
                    )

        raise KeyError(
            f"Failed to delete component with ID {component_id}: " +
            "no such component"
        )

    def clear_children(self, component_id: str) -> None:
        children = self.get_descendents(component_id)
        for child in children:
            try:
                self.updated = True
                for tree in self.tree_branches:
                    if not tree.freeze:
                        tree.components.pop(child.id, None)

            except UIError:
                logger = logging.getLogger("writer")
                logger.warning(
                    f"Failed to remove child with ID '{child.id}' " +
                    f"from component with ID '{component_id}': " +
                    "child is a builder-managed component.")
                # This might result in multiple consecutive warnings
                # for the same parent component, but we have to avoid "break"ing
                # due to that the component might still have CMC children


    def to_dict(self) -> Dict:
        trees = reversed(self.tree_branches)
        components = {}
        for tree in trees:
            components.update(tree.to_dict())

        return components

    def next_page_id(self) -> str:
        return f"page-{self.page_counter}"

    def fetch_updates(self):
        if self.updated:
            self.updated = False
            return self.to_dict()

        return

    def determine_position(self, parent_id: str, is_positionless: bool) -> int:
        if is_positionless:
            return -2

        children = self._get_direct_descendents(parent_id)
        cmc_children = list(filter(lambda c: c.isCodeManaged is True, children))
        if len(cmc_children) > 0:
            position = \
                max([0, max([child.position for child in cmc_children]) + 1])
            return position
        else:
            return 0

    def get_descendents(self, parent_id: str) -> List[Component]:
        children = self._get_direct_descendents(parent_id)
        desc = children.copy()
        for child in children:
            desc += self.get_descendents(child.id)

        return desc

    def branch(self, branch_id: Branch) -> ComponentTreeBranch:
        _branch = self._tree_branch(branch_id)
        if _branch is None:
            raise ValueError(f"Component tree with ID {branch_id} does not exist")

        return _branch

    def exists(self, branch_id: Branch) -> bool:
        """
        Checks if a component with the given ID exists in the component tree.
        """
        branch = self._tree_branch(branch_id)
        return branch is not None

    def is_frozen(self, branch_id: Branch) -> bool:
        """
        Checks if a component tree branch is frozen.
        """
        branch = self._tree_branch(branch_id)
        if branch is None:
            raise ValueError(f"Component tree with ID {branch_id} does not exist")

        return branch.freeze

    def _get_direct_descendents(self, parent_id: str) -> List[Component]:
        _all_components = self.components.values()
        children = list(filter(lambda c: c.parentId == parent_id, _all_components))
        return children

    def _tree_branch(self, branch: Optional[Branch]) -> Optional[ComponentTreeBranch]:
        if branch is None:
            return self.tree_branches[0]

        for tree in self.tree_branches:
            if branch == tree.component_tree_id:
                return tree

        return None


    def get_parent(self, component_id: str) -> List[str]:
        """
        Returns the list of parents, from the first to the highest level (root normally)

        :param component_id:
        :return:
        """
        components = self.components.values()
        parents = []
        current_node: Optional[str] = component_id
        while current_node is not None:
            for component in components:
                if component.id == current_node:
                    if component.parentId is not None:
                        parents.append(component.parentId)

                    current_node = component.parentId

        return parents


def build_base_component_tree() -> ComponentTree:
    """
    Create the base component tree. This tree is used when loading Writer Framework.

    It contains the components in common between all users.
    """
    bmc_tree_branch = ComponentTreeBranch(Branch.bmc)
    bmc_tree_branch.attach(Component(id="root", type="root", content={}))
    cmc_tree_branch = ComponentTreeBranch(Branch.initial_cmc)
    return ComponentTree([cmc_tree_branch, bmc_tree_branch])


def build_session_component_tree(base_component_tree: ComponentTree) -> ComponentTree:
    """
    Creates a session component tree.

    The session component tree is associated with a user session, i.e. a browser tab.
    If the user refreshes the page, a new component tree for the session is created.

    The builder managed component, copy from base component tree, can not be modified anymore.
    The code managed component can be modified.
    """
    session_tree_branch = ComponentTreeBranch(Branch.session_cmc)

    cmc_tree_branch = base_component_tree.branch(Branch.initial_cmc).clone()
    bmc_tree_branch = base_component_tree.branch(Branch.bmc).clone()
    # bmc_tree_branch.freeze = True

    return ComponentTree([session_tree_branch, cmc_tree_branch, bmc_tree_branch])


def ingest_bmc_component_tree(component_tree: ComponentTree, components: Dict[str, Any], ignore_freeze: bool = False):
    """
    Updates the builder managed component tree branch with the provided components.
    This method is used on the event `componentUpdate`.

    >>> ingest_bmc_component_tree(component_tree, {"root": {"type": "root", "content": {}})
    """
    assert component_tree.exists(Branch.bmc) is True, \
        "bmc component tree branch does not exists in this component tree"
    assert component_tree.is_frozen(Branch.bmc) is False or ignore_freeze is True, \
        "builder managed component tree are frozen and cannot be updated"

    component_tree.ingest(components, tree=Branch.bmc)

def filter_components_by(components: Dict[str, ComponentDefinition], parent: Optional[str] = None) -> Dict[str, ComponentDefinition]:
    """
    Filters tree components whose parent is {parent_id}, parent included

    >>> filtered_components = filter_components_by(components, parent="6a490318-239e-4fe9-a56b-f0f33d628c87")
    """

    if parent is not None:
        parent_list = {parent}
        target_components: Dict[str, ComponentDefinition] = {}
        has_changed = True
        # in some cases, the order requires to loops few times
        while True:
            if has_changed is False:
                break

            has_changed = False
            for c in components.values():
                if c['id'] == parent and c['id'] not in target_components:
                    target_components[c['id']] = c
                elif c.get('parentId', None) in parent_list and c['id'] not in target_components:
                    target_components[c['id']] = c
                    parent_list.add(c['id'])
                    has_changed = True

        components = target_components

    return components


def cmc_components_list(component_tree: ComponentTree) -> list:
    """
    Returns the list of code managed components in the component tree.

    (use mainly for testing purposes)
    """
    return list(component_tree.branch(Branch.initial_cmc).components.values())


def session_components_list(component_tree: ComponentTree) -> list:
    """
   Returns the list of session managed components in the component tree.

   (use mainly for testing purposes)
   """
    return list(component_tree.branch(Branch.session_cmc).components.values())

def export_component_tree(component_tree: ComponentTree, mode: ServeMode, only_update=False) -> Optional[Dict]:
    """
    Exports the component tree to the ui.

    >>> filtered_component_tree = core_ui.export_component_tree(session.session_component_tree, mode=writer.Config.mode)

    This function filters artifacts that should be hidden from the user, for example blueprints in run mode.

    :param component_tree: the full component tree
    :param mode: the mode of the application (edit, run)
    :param updated: return something only if component tree has been updated
    :return: a dictionary representing the component tree
    """
    if only_update is True and component_tree.updated is False:
        return None

    roots = ['root']
    if mode == "edit":
        roots.append('blueprints_root')

    _components: List[Component] = []
    for root in roots:
        _root_component = cast(Component, component_tree.get_component(root))
        _components.append(_root_component)
        _components += component_tree.get_descendents(root)

    # filters notes in run mode
    if mode == "run":
        _components = [c for c in _components if c.type != "note"]

    return {c.id: c.to_dict() for c in _components}

class UIError(Exception):
    ...


@contextlib.contextmanager
def use_component_tree(component_tree: ComponentTree):
    """
    Declares the component tree that will be manipulated during a context.

    The declared tree can be retrieved with the `current_component_tree` method.

    >>> with use_component_tree(component_tree):
    >>>     ui_manager = WriterUIManager()
    >>>     ui_manager.create_component("text", text="Hello, world!")

    :param component_tree:
    """
    token = _current_component_tree.set(component_tree)
    yield
    _current_component_tree.reset(token)


def current_component_tree() -> ComponentTree:
    """
    Retrieves the component tree of the current context or the base
    one if no context has been declared.

    :return:
    """
    tree = _current_component_tree.get()
    if tree is None:
        import writer.core
        return writer.core.base_component_tree

    return tree



import json
import logging
import os
import re
import sys
import tarfile
import tempfile
import time
from datetime import datetime, timedelta
from typing import List, Union

import click
import dateutil.parser
import pytz
import requests
from gitignore_parser import parse_gitignore

DEPLOY_TIMEOUT = int(os.getenv("WRITER_DEPLOY_TIMEOUT", 20))

@click.group()
def cloud():
    """A group of commands to deploy the app on writer cloud"""
    pass

@cloud.command()
@click.option('--api-key',
    default=lambda: os.environ.get("WRITER_API_KEY", None),
    allow_from_autoenv=True,
    show_envvar=True,
    envvar='WRITER_API_KEY',
    prompt="Enter your API key",
    hide_input=True, help="Writer API key"
)
@click.option('--env', '-e', multiple=True, default=[], help="Set environment variables for the app (e.g., --env KEY=VALUE)")
@click.option('--force', '-f', default=False, is_flag=True, help="Ignores warnings and overwrites the app")
@click.option('--verbose', '-v', default=False, is_flag=True, help="Enable verbose mode")
@click.argument('path')
def deploy(path, api_key, env, verbose, force):
    """Deploy the app from PATH folder."""

    deploy_url = os.getenv("WRITER_DEPLOY_URL", "https://api.writer.com/v1/deployment/apps")
    sleep_interval = int(os.getenv("WRITER_DEPLOY_SLEEP_INTERVAL", '5'))

    if not force:
        check_app(deploy_url, api_key)

    abs_path = os.path.abspath(path)
    if not os.path.isdir(abs_path):
        raise click.ClickException("A path to a folder containing a Writer Framework app is required. For example: writer cloud deploy my_app")

    env = _validate_env_vars(env)
    tar = pack_project(abs_path)
    try:
        upload_package(deploy_url, tar, api_key, env, verbose=verbose, sleep_interval=sleep_interval)
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 401:
            unauthorized_error()
        else:
            on_error_print_and_raise(e.response, verbose=verbose)
    except Exception as e:
        print(e)
        print("Error deploying app")
        sys.exit(1)
    finally:
        tar.close()

def _validate_env_vars(env: Union[List[str], None]) -> Union[List[str], None]:
    if env is None:
        return None
    for var in env:
        regex = r"^[a-zA-Z_]+[a-zA-Z0-9_]*=.*$"
        if not re.match(regex, var):
            logging.error(f"Invalid environment variable: {var}, please use the format ENV_VAR=value")
            sys.exit(1)
    return env

@cloud.command()
@click.option('--api-key',
    default=lambda: os.environ.get("WRITER_API_KEY", None),
    allow_from_autoenv=True,
    show_envvar=True,
    envvar='WRITER_API_KEY',
    prompt="Enter your API key",
    hide_input=True, help="Writer API key"
)
@click.option('--verbose', '-v', default=False, is_flag=True, help="Enable verbose mode")
def undeploy(api_key, verbose):
    """Stop the app, app would not be available anymore."""
    try:
        print("Undeploying app")
        deploy_url = os.getenv("WRITER_DEPLOY_URL", "https://api.writer.com/v1/deployment/apps")
        with requests.delete(deploy_url, headers={"Authorization": f"Bearer {api_key}"}) as resp:
            on_error_print_and_raise(resp, verbose=verbose)
            print("App undeployed")
            sys.exit(0)
    except Exception as e:
        print("Error undeploying app")
        print(e)
        sys.exit(1)

@cloud.command()
@click.option('--api-key',
    default=lambda: os.environ.get("WRITER_API_KEY", None),
    allow_from_autoenv=True,
    show_envvar=True,
    envvar='WRITER_API_KEY',
    prompt="Enter your API key",
    hide_input=True, help="Writer API key"
)
@click.option('--verbose', '-v', default=False, is_flag=True, help="Enable verbose mode")
def logs(api_key, verbose):
    """Fetch logs from the deployed app."""

    deploy_url = os.getenv("WRITER_DEPLOY_URL", "https://api.writer.com/v1/deployment/apps")
    sleep_interval = int(os.getenv("WRITER_DEPLOY_SLEEP_INTERVAL", '5'))

    try: 
        build_time = datetime.now(pytz.timezone('UTC')) - timedelta(days=4)
        start_time = build_time
        while True:
            prev_start = start_time
            end_time = datetime.now(pytz.timezone('UTC'))
            data = get_logs(deploy_url, api_key, {
                "buildTime": build_time,
                "startTime": start_time,
                "endTime": end_time,
            }, verbose=verbose)
            # order logs by date and print
            logs = data['logs']
            for log in logs:
                start_time = start_time if start_time > log[0] else log[0]
            if start_time == prev_start:
                start_time = datetime.now(pytz.timezone('UTC'))
                time.sleep(sleep_interval)
                continue
            for log in logs:
                print(log[0], log[1])
            print(start_time)
            time.sleep(sleep_interval)
    except Exception as e:
        print(e)
        sys.exit(1)
    except KeyboardInterrupt:
        sys.exit(0)

def pack_project(path):
    print(f"Creating deployment package from path: {path}")

    files = []
    def match(file_path) -> bool: return False
    if os.path.exists(os.path.join(path, ".gitignore")):
        match = parse_gitignore(os.path.join(path, ".gitignore"))
    for root, dirs, filenames in os.walk(path, followlinks=False):
        for filename in filenames:
            is_symlink = os.path.islink(os.path.relpath(os.path.join(root, filename), path))
            if is_symlink:
                print(f"[WARNING] Ignoring symlink: {os.path.relpath(os.path.join(root, filename), path)}")
                continue
            if "__pycache__" in root.split(os.path.sep):
                continue
            if ".git" in root.split(os.path.sep):
                continue
            if root == path and filename == "Dockerfile":
                print("[WARNING] Dockerfile found in project root. This will be ignored in the deployment package.")
                continue
            if root == path and filename == "service_entrypoint.py":
                print("[WARNING] service_entrypoint.py is a reserved file name. This will be ignored in the deployment package.")
                continue
            if not match(os.path.join(root, filename)):
                files.append(os.path.relpath(os.path.join(root, filename), path))

    f = tempfile.TemporaryFile(suffix='.tar')

    with tarfile.open(fileobj=f, mode="w") as tar:
        for file in files:
            print("Packing file:", file)
            tar.add(os.path.join(path, file), file)
    f.flush()

    return f

def check_app(deploy_url, token):
    url = _get_app_url(deploy_url, token)
    if url:
        print("[WARNING] This token was already used to deploy a different app")
        print(f"[WARNING] URL: {url}")
        print("[WARNING] If looking to deploy to a different URL, use a different API key. ")
        if input("[WARNING] Are you sure you want to overwrite? (y/N)").lower() != "y":
            sys.exit(1)

def _get_app_url(deploy_url: str, token: str) -> Union[str, None]:
    with requests.get(deploy_url, params={"lineLimit": 1}, headers={"Authorization": f"Bearer {token}"}) as resp:
        try:
            resp.raise_for_status()
        except Exception as e:
            print(e)
            print(resp.json())
            return None
        data = resp.json()
    return data['status']['url']

def get_logs(deploy_url, token, params, verbose=False):
    with requests.get(deploy_url, params = params, headers={"Authorization": f"Bearer {token}"}) as resp:
        on_error_print_and_raise(resp, verbose=verbose)
        data = resp.json()

        logs = []
        for log in data.get("logs", []):
            meta, msg = log["log"].split(" F ", 1)
            date, _ = meta.split(" ", 1)
            parsed_date = dateutil.parser.parse(date)
            logs.append((parsed_date, msg))
        logs.sort(key=lambda x: x[0])
        return {"status": data["status"], "logs": logs}

def check_service_status(deploy_url, token, build_id, build_time, start_time, end_time, last_status):
    data = get_logs(deploy_url, token, {
        "buildId": build_id,
        "buildTime": build_time,
        "startTime": start_time,
        "endTime": end_time
    })
    if data["status"]["status"] == "DEPLOYING" and data["status"]["status"] != last_status:
        print("Build completed. Deploying app...")
    status = data["status"]["status"]
    url = data["status"]["url"]
    for log in data.get("logs", []):
        print(log[0], log[1])
    return status, url

def dictFromEnv(env: List[str]) -> dict:
    env_dict = {}
    if env is None:
        return env_dict
    for e in env:
        key, value = e.split("=", 1)
        env_dict[key] = value
        print('Environment variable:', key)

    return env_dict


def upload_package(deploy_url, tar, token, env, verbose=False, sleep_interval=5):
    print("Uploading package to deployment server")
    tar.seek(0)
    files = {'file': tar}
    start_time = datetime.now(pytz.timezone('UTC'))
    build_time = start_time

    with requests.post(
        url = deploy_url, 
        headers = {
            "Authorization": f"Bearer {token}",
        },
        files=files,
        data={"envs": json.dumps(dictFromEnv(env))}
    ) as resp:
        on_error_print_and_raise(resp, verbose=verbose)
        data = resp.json()
        build_id = data["buildId"]

    print("Package uploaded. Building...")
    status = "WAITING"
    url = ""
    while status not in ["COMPLETED", "FAILED"] and datetime.now(pytz.timezone('UTC')) < build_time + timedelta(minutes=DEPLOY_TIMEOUT):
        end_time = datetime.now(pytz.timezone('UTC'))
        status, url = check_service_status(deploy_url, token, build_id, build_time, start_time, end_time, status)
        time.sleep(sleep_interval)
        start_time = end_time

    if status == "COMPLETED":
        print("Deployment successful")
        print(f"URL: {url}")
        sys.exit(0)
    elif status == "FAILED":
        print("Deployment failed")
        print(f"URL: {url}")
        sys.exit(1)
    else:
        time.sleep(sleep_interval)
        check_service_status(deploy_url, token, build_id, build_time, start_time, datetime.now(pytz.timezone('UTC')), status)
        print("Deployment status timeout")
        print(f"URL: {url}")
        sys.exit(2)

def on_error_print_and_raise(resp, verbose=False):
    try:
        resp.raise_for_status()
    except Exception as e:
        if verbose:
            print(resp.json())
        raise e

def unauthorized_error():
    print("Unauthorized. Please check your API key.")
    sys.exit(1)




import json

from writer.abstract import register_abstract_template
from writer.blocks.base_block import BlueprintBlock
from writer.ss_types import AbstractTemplate


class HTTPRequest(BlueprintBlock):
    @classmethod
    def register(cls, type: str):
        super(HTTPRequest, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="blueprints_node",
                writer={
                    "name": "HTTP request",
                    "description": "Sends an HTTP request to an API endpoint. Used to fetch or send data.",
                    "category": "Other",
                    "fields": {
                        "method": {
                            "name": "Method",
                            "type": "Text",
                            "options": {
                                "GET": "GET",
                                "POST": "POST",
                                "PUT": "PUT",
                                "PATCH": "PATCH",
                                "DELETE": "DELETE",
                            },
                            "validator": {
                                "type": "string",
                                "enum": ["GET", "POST", "PUT", "PATCH", "DELETE", ""],
                            },
                        },
                        "url": {
                            "name": "URL",
                            "type": "Text",
                            "control": "Textarea",
                        },
                        "headers": {
                            "name": "Headers",
                            "type": "Key-Value",
                            "default": "{}",
                            "validator": {
                                "type": "object",
                                "patternProperties": {
                                    "^.*$": {
                                        "type": ["string", "number", "boolean"],
                                    },
                                },
                                "additionalProperties": True,
                            },
                        },
                        "bodyType": {
                            "name": "Body type",
                            "type": "Text",
                            "description": "Specify whether to interpret the body as plain text or JSON.",
                            "options": {
                                "text": "Plain text",
                                "JSON": "JSON",
                            },
                            "default": "text",
                        },
                        "body": {"name": "Body", "type": "Text", "control": "Textarea"},
                    },
                    "outs": {
                        "success": {
                            "name": "Success",
                            "description": "The request was successful.",
                            "style": "success",
                        },
                        "responseError": {
                            "name": "Response error",
                            "description": "The connection was established successfully but an error response code was received or the response was invalid.",
                            "style": "error",
                        },
                        "connectionError": {
                            "name": "Connection error",
                            "description": "The connection couldn't be established.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def run(self):
        try:
            method = self._get_field("method", False, "GET")
            url = self._get_field("url")
            headers = self._get_field("headers", True, default_field_value="{}")
            body_type = self._get_field("bodyType")
            body = None
            raw_body = None

            if body_type == "JSON":
                body = self._get_field("body", as_json=True)
                raw_body = json.dumps(body)
                headers.setdefault("Content-Type", "application/json")
            else:
                body = self._get_field("body", as_json=False)
                raw_body = body

            with self.acquire_httpx_client() as client:
                res = client.request(method, url, headers=headers, content=raw_body, timeout=180)

                content_type = res.headers.get("Content-Type", "")
                is_response_json = "application/json" in content_type

                self.result = {
                    "request": {
                        "url": str(res.request.url),
                        "headers": dict(res.request.headers),
                        "body": res.request.content.decode("utf-8", errors="replace"),
                    },
                    "headers": dict(res.headers),
                    "status_code": res.status_code,
                    "body": res.json() if is_response_json else res.text,
                }

                if res.is_success:
                    self.outcome = "success"
                else:
                    self.outcome = "responseError"
                    raise RuntimeError(
                        f"HTTP response with code {res.status_code} and message {res.text}"
                    )

        except json.JSONDecodeError as e:
            self.outcome = "responseError"
            raise e
        except Exception as e:
            if self.outcome is None or self.outcome not in ("responseError",):
                self.outcome = "connectionError"
            raise e



import json
import re

from writer.abstract import register_abstract_template
from writer.blocks.base_block import WriterBlock
from writer.ss_types import AbstractTemplate


class WriterClassification(WriterBlock):
    @classmethod
    def register(cls, type: str):
        super(WriterClassification, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="blueprints_node",
                writer={
                    "name": "Classification",
                    "description": "Classifies text into predefined categories using AI. Useful for tagging and routing inputs.",
                    "category": "Writer",
                    "fields": {
                        "text": {
                            "name": "Text",
                            "type": "Text",
                            "desc": "The text you want to classify.",
                        },
                        "categories": {
                            "name": "Categories",
                            "type": "Key-Value",
                            "default": "{}",
                            "desc": "The keys should be the categories you want to classify the text into, for example 'valid' and 'invalid', and the values are the criteria for each category. Category names should contain only letters of the English alphabet, digits, underscores, and spaces.",
                        },
                        "additionalContext": {
                            "name": "Additional context",
                            "type": "Text",
                            "control": "Textarea",
                            "desc": "Any additional information that might help the AI in making the classification decision.",
                        },
                    },
                    "outs": {
                        "category": {"field": "categories", "style": "dynamic"},
                        "error": {
                            "name": "Error",
                            "description": "There was an error classifying the text.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def run(self):
        try:
            import writer.ai

            text = self._get_field("text", required=True)
            additional_context = self._get_field("additionalContext")
            categories = self._get_field("categories", as_json=True, required=True)
            conversation = writer.ai.Conversation()

            invalid_categories = [category for category in categories if not re.fullmatch(r"[\w ]+", category, flags=re.ASCII)]
            if invalid_categories:
                self.outcome = "error"
                raise ValueError(f"Category names should contain only letters of the English alphabet, digits, underscores and spaces. Invalid categories: {', '.join(invalid_categories)}")

            config = {}

            prompt = f"""
Classify the text under “CONTENT” into one of the following categories:

{ json.dumps(categories) }

Your output should be only the key and not contain anything else. For example: { " , ".join(list(categories.keys())) }.

Additional context:

{ additional_context }

CONTENT:
------
{ text }
"""
            
            response_format = {
                "type": "json_schema",
                "json_schema": {
                    "schema": {
                        "type": "string",
                        "enum": list(categories.keys())
                    }
                }
            }

            conversation += {
                "role": "user",
                "content": prompt,
            }

            msg = conversation.complete(response_format=response_format, config=config)
            raw_content = msg.get("content")
            if not raw_content:
                self.outcome = "error"
                raise RuntimeError("No content returned from the model. Please validate the prompt and model configuration.")

            try:
                # Attempt to parse the raw content as JSON
                category_result = json.loads(raw_content)
            except json.JSONDecodeError:
                self.outcome = "error"
                raise RuntimeError(
                    f"Failed to decode JSON content. The raw content was: {raw_content}. Please validate the prompt and model configuration."
                )

            self.result = category_result
            self.outcome = f"category_{category_result}"
        except BaseException as e:
            self.outcome = "error"
            raise e



import json

from writer.abstract import register_abstract_template
from writer.blocks.base_block import WriterBlock
from writer.ss_types import AbstractTemplate

DEFAULT_MODEL = "palmyra-x5"


class WriterStructuredOutput(WriterBlock):
    @classmethod
    def register(cls, type: str):
        super(WriterStructuredOutput, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="blueprints_node",
                writer={
                    "name": "Structured output",
                    "description": "Defines a JSON response format that the agent will use to structure its output.",
                    "category": "Writer",
                    "fields": {
                        "prompt": {"name": "Prompt", "type": "Text", "control": "Textarea", "desc": "Description of a JSON object to be created."},
                        "modelId": {
                            "name": "Model",
                            "type": "Model Id",
                            "default": DEFAULT_MODEL
                        },
                        "jsonSchema": {
                            "name": "JSON Schema",
                            "desc": "JSON schema that defines the structure of the response. For example, `{\"type\": \"object\", \"properties\": {...}, \"required\": [...]}`.",
                            "type": "JSON",
                            "default": "{}"
                        },
                        "max_tokens": {
                            "name": "Max output tokens",
                            "type": "Number",
                            "default": "1024",
                            "validator": {
                                "type": "number",
                                "minimum": 1,
                                "maximum": 16384,
                            }
                        }
                    },
                    "outs": {
                        "success": {
                            "name": "Success",
                            "description": "The structured output was generated.",
                            "style": "success",
                        },
                        "error": {
                            "name": "Error",
                            "description": "There was an error generating the output.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def run(self):
        import writer.ai

        try:
            prompt = self._get_field("prompt")
            model_id = self._get_field("modelId", False, default_field_value=DEFAULT_MODEL)
            conversation = writer.ai.Conversation()
            schema = self._get_field("jsonSchema", True, default_field_value="{}")
            max_tokens = int(self._get_field("max_tokens", False, "1024"))

            response_format = {
                "type": "json_schema",
                "json_schema": {
                    "schema": schema
                    }
                }

            conversation += {
                "role": "user",
                "content": prompt,
            }
            config = { "model": model_id, "max_tokens": max_tokens }
            msg = conversation.complete(response_format=response_format, config=config)
            conversation += msg

            raw_content = msg.get("content")
            if not raw_content:
                self.outcome = "error"
                raise RuntimeError("No content returned from the model. Please validate the prompt and model configuration.")

            try:
                # Attempt to parse the raw content as JSON
                content = json.loads(raw_content)
            except json.JSONDecodeError:
                self.outcome = "error"
                raise RuntimeError(
                    f"Failed to decode JSON content. The raw content was: {raw_content}. Please validate the prompt and model configuration."
                )

            self.result = content
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



import logging
import os
from functools import partial
from typing import Any, Dict, List, Literal, Optional, Protocol

import httpx

logger = logging.getLogger("kv_storage")


class _WrappedRequestFunc(Protocol):
    def __call__(
        self,
        headers: Dict[str, str],
        timeout: int,
    ) -> httpx.Response: ...


class KeyValueStorage:
    def __init__(self, client: Optional[httpx.Client] = None) -> None:
        base_url = os.getenv("WRITER_BASE_URL")
        self.api_key = os.getenv("WRITER_API_KEY")
        if None in (base_url, self.api_key):
            logger.warning("Missing required environment variables for KV storage access")
        self.api_url = f"{base_url}/v1" if base_url else None

        self._client = client if client is not None else httpx

    def _get_agent_ids(self):
        from writer.core import get_session
        current_session = get_session()

        if current_session:
            headers = current_session.headers or {}
            agent_id = headers.get("x-agent-id") or os.getenv("WRITER_APP_ID")
            org_id = headers.get("x-organization-id") or os.getenv("WRITER_ORG_ID")
            return (agent_id, org_id)

        agent_id = os.getenv("WRITER_APP_ID")
        org_id = os.getenv("WRITER_ORG_ID")
        return (agent_id, org_id)

    def get(self, key: str, type_: Literal["data", "secret"]) -> Dict[str, Any]:
        return self._request(partial(self._client.get, url=f"{self.api_url}/agent_{type_}/{key}")).json()

    def get_data_keys(self) -> List[str]:
        return self._request(partial(self._client.get, url=f"{self.api_url}/agent_data")).json()["keys"]
    
    def save(self, key: str, data: Any) -> Dict[str, Any]:
        try:
            return self._create(key, data).json()
        except httpx.HTTPStatusError as e:
            if "already exists" in e.response.text:
                return self._update(key, data).json()
            raise e

    def _create(self, key: str, data: Any) -> httpx.Response:
        return self._request(partial(self._client.post, url=f"{self.api_url}/agent_data", json={"key": key, "data": data}))

    def _update(self, key: str, data: Any) -> httpx.Response:
        return self._request(partial(self._client.put, url=f"{self.api_url}/agent_data/{key}", json={"data": data}))

    def delete(self, key: str) -> Dict[str, str]:
        self._request(partial(self._client.delete, url=f"{self.api_url}/agent_data/{key}"))
        return {"key": key}

    def _request(self, request_func: _WrappedRequestFunc) -> httpx.Response:

        agent_id, org_id = self._get_agent_ids()
        if None in (agent_id, org_id):
            raise ValueError("Can't access KV storage. Missing agent id or org id")

        if None in (self.api_key, self.api_url):
            raise ValueError("Can't access KV storage. Missing required env vars")

        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "X-Organization-Id": org_id,
            "X-Agent-Id": agent_id,
        }

        response = request_func(headers=headers, timeout=3)
        response.raise_for_status()
        return response

    def is_accessible(self) -> bool:
        if None in self._get_agent_ids():
            return False
        if None in (self.api_key, self.api_url):
            return False
        return True

writer_kv_storage = KeyValueStorage()
